﻿using AirFusion.Manager.DeviceManager;
using AirFusion.WebService.Adapters;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace AirFusion.ConnectionManager
{
	internal class CheckRemoteDeviceStatusManager : ConnectionInterface
	{
		#region define some const value

		private static CheckRemoteDeviceStatusManager _pCheckRemoteDeviceStatusManager = null;
		private static object _pInstanceLock = new object();
		private ConnectDeviceStatusType connStatusType;

		#endregion define some const value

		#region public method

		public event EventHandler<ConnectEventArgs> ConnectionFlowEvent;

		public event EventHandler<AppEventArgs> ConnectStepEvent;

		/// <summary>
		/// For avoid CheckRemoteDeviceStatusManager double create.
		/// </summary>
		public static CheckRemoteDeviceStatusManager Instance
		{
			get
			{
				if (_pCheckRemoteDeviceStatusManager == null)
				{
					lock (_pInstanceLock)
					{
						if (_pCheckRemoteDeviceStatusManager == null) _pCheckRemoteDeviceStatusManager = new CheckRemoteDeviceStatusManager();
					}
				}
				return _pCheckRemoteDeviceStatusManager;
			}
		}

		private CheckRemoteDeviceStatusManager()
		{
			this.connStatusType = ConnectDeviceStatusType.None;
		}

		public void Connect(AppDeviceInfo deviceInfo, FlowControlManager controlmanager)
		{
			LogtoFile.TrackEnter();

			controlmanager.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.PhoneWakeUp));

			//Reset connStatusType to default value
			this.connStatusType = ConnectDeviceStatusType.None;

			///First Need to Wakeup remote device and need to wait the task to complete to check wakeup result
			WakeupRemoteDevice(deviceInfo).ContinueWith(delegate(Task<ConnectDeviceStatusType> Wakeuptask)
			{
				this.connStatusType = Wakeuptask.Result;

				///Check Wakeup remote device is success, if Wakeup remote device is success,
				///the value is CheckConnectionStatusType.WakeupSuccess
				///
				///Error Result : CheckConnectionStatusType.WakeupFail, CheckConnectionStatusType.ConfirmIPFail
				///
				///Need to Double Check Error Result : CheckConnectionStatusType.DandelionFail
				if (Wakeuptask.Result != ConnectDeviceStatusType.WakeupSuccess)
				{
					LogtoFile.TrackLeave();
					this.Disconnect(deviceInfo);
				}
				else
				{
					///When Wakeup remote device success, we need to check remote device status is paired/unpaired
					///and need to wait the task to complete to check remote device status result
					GetRemoteDeviceStatus(deviceInfo).ContinueWith(delegate(Task<ConnectDeviceStatusType> DeviceStatustask)
					{
						this.connStatusType = DeviceStatustask.Result;

						///Check remote device status is Paired or not, if check remote device is paired,
						///the value is CheckConnectionStatusType.CheckDeviceStatusPair
						///
						///Error Result : CheckConnectionStatusType.CheckDeviceStatusUnPair,
						///               CheckConnectionStatusType.CheckDeviceStatusTimeout
						///
						///Need to Double Check Error Result : CheckConnectionStatusType.CheckDeviceStatusFail
						if (DeviceStatustask.Result != ConnectDeviceStatusType.CheckDeviceStatusPair)
						{
							LogtoFile.TrackLeave();
							this.Disconnect(deviceInfo);
						}
						else
						{
							///When check remote device status is paired, we need to get remote device info
							///and need to wait the task to complete to check remote device info result
							GetRemoteDeviceInfo(deviceInfo).ContinueWith(delegate(Task<ConnectDeviceStatusType> DeviceInfotask)
							{
								this.connStatusType = DeviceInfotask.Result;

								///Check get remote device info is success, if get remote device info is success,
								///the value is CheckConnectionStatusType.GetDeviceInfoSuccess
								///
								///Error Result : CheckConnectionStatusType.GetDeviceInfoTimeout
								///
								///Need to Double Check Error Result : CheckConnectionStatusType.GetDeviceInfoFail
								if (DeviceInfotask.Result != ConnectDeviceStatusType.GetDeviceInfoSuccess)
								{
									LogtoFile.TrackLeave();
									this.Disconnect(deviceInfo);
								}
								else
								{
									LogtoFile.TrackLeave();
									this.Connected();
								}
							}).ConfigureAwait(continueOnCapturedContext: false);
						}
					}).ConfigureAwait(continueOnCapturedContext: false);
				}
			}).ConfigureAwait(continueOnCapturedContext: false);
		}

		/// <summary>
		/// When connction finish, need to create a call back funciton to info the result
		/// </summary>
		public void Connected()
		{
			this.SendConnectionEvent(new ConnectEventArgs(ConnectionEventType.CheckRemoteDeviceStatusSuccess));
		}

		/// <summary>
		/// The Function will try to disconnecting about connction work. Because in the Manager is handle Check
		/// remote device status manager, so we don't disconnction any work. Just call disconncted function.
		/// </summary>
		public void Disconnect(AppDeviceInfo deviceInfo, bool IsAuto = true)
		{
			LogtoFile.TrackEnter();
			this.Disconnected(IsAuto);
			LogtoFile.TrackLeave();
		}

		/// <summary>
		/// When Disconnected finished, need to crate a call back function to info the result.
		/// </summary>
		/// <param name="deviceInfoStatus"></param>
		public void Disconnected(bool IsAuto)
		{
			LogtoFile.TrackEnter();
			//Need to send event to info UI about disconnected is success.
			if (IsAuto)
				this.SendConnectionEvent(new ConnectEventArgs(ConnectionEventType.CheckRemoteDeviceStatusFail, this.connStatusType));
			LogtoFile.TrackLeave();
		}

		#endregion public method

		#region private method

		#region Remote Device WakeUp

		/// <summary>
		/// Wakeup the remote device asynchronously by WebService.
		/// </summary>
		/// <param name="deviceInfo"></param>
		/// <returns>
		/// If return wakeup success represent the remote device has been wakeup
		/// If return wakeup fail represent the remote device can't be wakeup(maybe IP has some error)
		/// If reutrn DandelionFail represent webservice try to get IP from dandelion has some issue. Need to Double check.
		/// If return ConfirmIPFail represent webservice try to get IP success but can't "ping" the ip maybe the remote device IP has been changed.
		/// </returns>
		private async Task<ConnectDeviceStatusType> WakeupRemoteDevice(AppDeviceInfo deviceInfo)
		{
			ConnectDeviceStatusType result = ConnectDeviceStatusType.WakeupFail;

			#region create task for device wakeup by origin IP and Reconfirm IP(By dandelion or mDNS update)

			try
			{
				LogtoFile.TrackEnter();

				List<Task<ConnectDeviceStatusType>> tasks = new List<Task<ConnectDeviceStatusType>>();

				//The task will update IP by Dandelion and mDNS and wakeup remote device
				tasks.Add(this.DoWakeUpRemoteDevice(deviceInfo, true));

				//The task will use original IP to wakeup remote device
				tasks.Add(this.DoWakeUpRemoteDevice(deviceInfo, false));

				while (true)
				{
					if (tasks.Count == 0)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "No any task need to be waited!!", "");
						break;
					}

					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Wait for any task completed", "");

					var task = await Task.WhenAny(tasks);

					result = task.Result;
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Get result: " + result, "");

					tasks.Remove(task);

					if (result == ConnectDeviceStatusType.WakeupSuccess)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "WakeUp Success no need to wait any task", "");
						break;
					}
					else if (result == ConnectDeviceStatusType.VersionNotMatch)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "WakeUp fail because version not match", "");
						break;
					}
					else
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "Wake up fail, need to wait last task", "");
					}
				}
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
				return result;
			}
			finally
			{
				LogtoFile.TrackLeave();
			}

			#endregion create task for device wakeup by origin IP and Reconfirm IP(By dandelion or mDNS update)

			return result;
		}

		/// <summary>
		/// Wakeup the remote device asynchronously by WebService.
		/// </summary>
		/// <param name="deviceInfo"></param>
		/// <returns>
		/// If return wakeup success represent the remote device has been wakeup
		/// If return wakeup fail represent the remote device can't be wakeup(maybe IP has some error)
		/// If reutrn DandelionFail represent webservice try to get IP from dandelion has some issue. Need to Double check.
		/// If return ConfirmIPFail represent webservice try to get IP success but can't "ping" the ip maybe the remote device IP has been changed.
		/// </returns>
		private async Task<ConnectDeviceStatusType> DoWakeUpRemoteDevice(AppDeviceInfo deviceInfo, bool isDoReconfirmIP)
		{
			try
			{
				LogtoFile.TrackEnter();

				//Reconfirm IP address by WebService

				#region Reconfirm IP Address

				if (isDoReconfirmIP)
				{
					ReconfirmAddressResult reconfirmIPResult = null;

					try
					{
						reconfirmIPResult = await this.WebServiceReconfirmAddressAsync(deviceInfo.Device).ConfigureAwait(false);

						if (reconfirmIPResult != null)
						{
							deviceInfo.ReconfirmAddressResult = reconfirmIPResult;

							if (reconfirmIPResult.Addresses == null)
							{
								return ConnectDeviceStatusType.DandelionFail;
							}
							else if (reconfirmIPResult.Addresses.Count == 0)
							{
								return ConnectDeviceStatusType.DandelionFail;
							}

							if (reconfirmIPResult.FirstConnectableAddress != null)
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, ">>>>>> reconfirmed IP: " + reconfirmIPResult.FirstConnectableAddress, "");

								LogtoFile.WriteLog("", Configs.DebugLevelValue, "deviceInfo.ServerIP: " + deviceInfo.ServerIP, "");
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "deviceInfo.LastServerIP: " + deviceInfo.LastServerIP, "");

								if (!deviceInfo.ServerIP.Equals(deviceInfo.LastServerIP))
								{
									LogtoFile.WriteLog("", Configs.DebugLevelValue, "deviceInfo.ServerIP is different with deviceInfo.LastServerIP", "");
									deviceInfo.LastServerIP = reconfirmIPResult.FirstConnectableAddress.ToString();
									LogtoFile.WriteLog("", Configs.DebugLevelValue, "Update LastServerIP", "");
								}
								else
								{
									LogtoFile.WriteLog("", Configs.DebugLevelValue, "deviceInfo.ServerIP is the same as deviceInfo.LastServerIP", "");
								}
							}
							else
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, ">>>>>> reconfirmIPResult.FirstConnectableAddress value is null", "");
								return ConnectDeviceStatusType.ConfirmIPFail;
							}
						}
						else
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, ">>>>>> reconfirmIPResult value is null", "");
							Debugger.Break();
							return ConnectDeviceStatusType.DandelionFail;
						}
					}
					catch (Exception ex)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
						return ConnectDeviceStatusType.DandelionFail;
					}
				}

				#endregion Reconfirm IP Address

				// Do wake up

				#region WakeUp Device

				bool isWakeUpSuccess = false;
				if (null != deviceInfo.Device)
				{
					try
					{
						isWakeUpSuccess = await WebServiceWakeUpRemoteDevice(deviceInfo.Device);
					}
					catch (WebServiceVersionNotMatchedException wex)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, wex.ToString(), "");
						throw wex;
					}
					catch (Exception ex)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
					}
				}
				else
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "deviceInfo.Device is null", "");
				}

				if (!isWakeUpSuccess)
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Failed to wake up !", "");

					//0412 Johnson add, when wakeup faile, we should not to do anything.
					return ConnectDeviceStatusType.WakeupFail;
				}
				else
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Wake up successfully!", "");
					return ConnectDeviceStatusType.WakeupSuccess;
				}

				#endregion WakeUp Device
			}
			catch (WebServiceVersionNotMatchedException wex)
			{
				FlowControlManager.Instance.RemoteDeviceapplicationVersion = wex.ApplicationVersion;
				FlowControlManager.Instance.RemoteDeviceminimumTargetVersion = wex.MinimumTargetVersion;

				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Wake up WebServiceVersionNotMatchedException ex : " + wex.ToString(), "");
				return ConnectDeviceStatusType.VersionNotMatch;
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Wake up exception ex : " + ex.ToString(), "");
				return ConnectDeviceStatusType.WakeupFail;
			}
			finally
			{
				LogtoFile.TrackLeave();
			}
		}

		#endregion Remote Device WakeUp

		#region Remote Device Status

		/// <summary>
		/// Requests the remote device pairing status asynchronously by WebService.
		/// </summary>
		/// <param name="deviceInfo"></param>
		/// <returns>The remote pairing status of the device by WebService.
		/// If return "DeviceInfoStatus.Connectable" represent we can connetion to this device.
		/// If return "DeviceInfoStatus.Unpair" represent we can't connection to this device.
		/// If return "DeviceInfoStatus.CheckDeviceStatusTimeout" represent timeout by webservice(maybe network has some issue)
		/// If return "DeviceInfoStatus.CheckDeviceStatusFail" represent WebService has some issue, need to double check.
		/// </returns>
		private async Task<ConnectDeviceStatusType> GetRemoteDeviceStatus(AppDeviceInfo deviceInfo)
		{
			ConnectDeviceStatusType result = ConnectDeviceStatusType.CheckDeviceStatusFail;

			#region Get DeviceStatus

			LogtoFile.TrackEnter();

			for (int i = 0; i < Configs.CheckRemoteDeviceStatusRetryTimes; i++)
			{
				// To get remote status
				FlowControlManager.Instance.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.GetDeviceStatus, i));
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Call RequestPairingStatus()", "");
				result = await this.DoGetRemoteDeviceStatus(deviceInfo.Device).ConfigureAwait(false);
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "status: " + result, "");

				if (result == ConnectDeviceStatusType.CheckDeviceStatusPair || result == ConnectDeviceStatusType.CheckDeviceStatusUnPair)
					break;

				LogtoFile.WriteLog("", Configs.DebugLevelValue, "GetRemoteDeviceStatus fail, wait 1s", "");
				//Thread.Sleep(Configs.CheckRemoteDeviceStatusRetryTimePeriod);
				await Task.Delay(Configs.CheckRemoteDeviceStatusRetryTimePeriod);
			}

			LogtoFile.TrackLeave();

			return result;

			#endregion Get DeviceStatus
		}

		/// <summary>
		/// Requests the remote device pairing status asynchronously by WebService.
		/// </summary>
		/// <param name="deviceInfo"></param>
		/// <returns>The remote pairing status of the device by WebService.
		/// If return "DeviceInfoStatus.Connectable" represent we can connetion to this device.
		/// If return "DeviceInfoStatus.Unpair" represent we can't connection to this device.
		/// If return "DeviceInfoStatus.CheckDeviceStatusTimeout" represent timeout by webservice(maybe network has some issue)
		/// If return "DeviceInfoStatus.CheckDeviceStatusFail" represent WebService has some issue, need to double check.
		/// </returns>
		private async Task<ConnectDeviceStatusType> DoGetRemoteDeviceStatus(AFUserDevice device)
		{
			LogtoFile.TrackEnter();

			// If the device isn't connectable
			try
			{
				AFUserDevice.PairingStatus originalPairingStatus = device.PairStatus;
				AFUserDevice.PairingStatus newPairingStatus = await this.WebServiceRequestPairingStatusAsync(device).ConfigureAwait(false);  // Request remote status

				if (newPairingStatus == AFUserDevice.PairingStatus.UnPaired)
				{
					return ConnectDeviceStatusType.CheckDeviceStatusUnPair;
				}
				else
				{
					return ConnectDeviceStatusType.CheckDeviceStatusPair;
				}
			}
			catch (TimeoutException)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "device.RequestPairingStatus() Timeout", "");
				return ConnectDeviceStatusType.CheckDeviceStatusTimeout;
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "device.RequestPairingStatus() error: " + ex.ToString(), "");
				return ConnectDeviceStatusType.CheckDeviceStatusFail;
			}
			finally
			{
				LogtoFile.TrackLeave();
			}
		}

		#endregion Remote Device Status

		#region Remote Device Info

		/// <summary>
		/// Requests the remote device Info asynchronously by WebService.
		/// </summary>
		/// <param name="deviceInfo"></param>
		/// <returns>The remote pairing status of the device by WebService.
		/// If return "DeviceInfoStatus.Connectable" represent we can connetion to this device.
		/// If return "DeviceInfoStatus.Unpair" represent we can't connection to this device.
		/// If return "DeviceInfoStatus.CheckDeviceStatusTimeout" represent timeout by webservice(maybe network has some issue)
		/// If return "DeviceInfoStatus.CheckDeviceStatusFail" represent WebService has some issue, need to double check.
		/// </returns>
		private async Task<ConnectDeviceStatusType> GetRemoteDeviceInfo(AppDeviceInfo deviceInfo)
		{
			ConnectDeviceStatusType result = ConnectDeviceStatusType.GetDeviceInfoFail;

			#region Get Device Info

			LogtoFile.TrackEnter();

			for (int i = 0; i < Configs.CheckRemoteDeviceInfoRetryTimes; i++)
			{
				FlowControlManager.Instance.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.GetDeviceInformation, i));

				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Call RequestInformationAsync()", "");
				result = await this.WebServiceRequestRemoteDeviceInfo(deviceInfo.Device).ConfigureAwait(false);
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "status: " + result, "");

				if (result == ConnectDeviceStatusType.GetDeviceInfoSuccess)
					return ConnectDeviceStatusType.GetDeviceInfoSuccess;

				LogtoFile.WriteLog("", Configs.DebugLevelValue, "RequestRemoteDeviceInfo fail, wait 1s", "");
				await Task.Delay(Configs.CheckRemoteDeviceInfoRetryTimePeriod);
			}

			LogtoFile.TrackLeave();

			return result;

			#endregion Get Device Info
		}

		#endregion Remote Device Info

		#region Call WebService to check information

		/// <summary>
		/// Call WebService funciton to update IP by dandelion or mDNS
		/// </summary>
		/// <param name="targetDevice"></param>
		/// <returns>ReconfirmAddressResult structure</returns>
		private async Task<ReconfirmAddressResult> WebServiceReconfirmAddressAsync(AFUserDevice targetDevice)
		{
			return await targetDevice.ReconfirmAddressAsync().ConfigureAwait(false);
		}

		/// <summary>
		/// Call WebService function to wakeup remote device
		/// </summary>
		/// <param name="targetDevice"></param>
		/// <returns>
		/// True represent the remote device has been wakeup success
		/// False represent the remote device can't be wakeup
		/// </returns>
		private async Task<bool> WebServiceWakeUpRemoteDevice(AFUserDevice targetDevice)
		{
			return await targetDevice.WakeUpAsync().ConfigureAwait(false);
		}

		/// <summary>
		/// Call WebService function to get remote device status
		/// </summary>
		/// <param name="targetDevice"></param>
		/// <returns>
		/// AFUserDevice.PairingStatus
		/// </returns>
		private async Task<AFUserDevice.PairingStatus> WebServiceRequestPairingStatusAsync(AFUserDevice targetDevice)
		{
			return await targetDevice.RequestPairingStatusAsync().ConfigureAwait(false);
		}

		/// <summary>
		/// Requests the remote device pairing status asynchronously by WebService.
		/// </summary>
		/// <param name="deviceInfo"></param>
		/// <returns>The remote pairing status of the device by WebService.
		/// If return "CheckConnectionStatusType.Connectable" represent we can get this device inforamtion.
		/// If return "CheckConnectionStatusType.GetDeviceInfoTimeout" represent timeout by webservice(maybe network has some issue)
		/// If return "CheckConnectionStatusType.GetDeviceInfoFail" represent WebService has some issue, need to double check.
		/// </returns>
		private async Task<ConnectDeviceStatusType> WebServiceRequestRemoteDeviceInfo(AFUserDevice device)
		{
			LogtoFile.TrackEnter();

			try
			{
				await device.RequestInformationAsync().ConfigureAwait(false);
				return ConnectDeviceStatusType.GetDeviceInfoSuccess;
			}
			catch (TimeoutException)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Request Remote Information Timout", "");
				return ConnectDeviceStatusType.GetDeviceInfoTimeout;
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
				return ConnectDeviceStatusType.GetDeviceInfoFail;
			}
			finally
			{
				LogtoFile.TrackLeave();
			}
		}

		#endregion Call WebService to check information

		#region Event Handle

		private void SendConnectionEvent(ConnectEventArgs args)
		{
			try
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Ready to send ConnectionEvent: " + args.EventType, "");

				if (null != this.ConnectionFlowEvent)
				{
					var receivers = this.ConnectionFlowEvent.GetInvocationList();
					foreach (EventHandler<ConnectEventArgs> receiver in receivers)
					{
						receiver.BeginInvoke(this, args, null, null);
						//receiver.Invoke(this, args);
					}
				}
				else
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "this.ConnectionFlowEvent is null", "");
				}
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
			}
		}

		#endregion Event Handle

		#endregion private method
	}
}