﻿using AirFusion.Common;
using InTheHand.Net;
using System;
using System.IO.Pipes;
using System.Runtime.InteropServices;

namespace AirFusion.PipeHelper
{
	internal class PipeManager<TClass>
		where TClass : class, new()
	{
		internal delegate void pipeMsgReceivedEventHandler(TClass data);

		internal pipeMsgReceivedEventHandler MsgReceived;

		public string PipeName { get; internal set; }

		public int PipeTimeout { get; internal set; }

		public PipeManager()
		{
			PipeName = "Pipe";
			PipeTimeout = 5000;
		}

		public void StartListen()
		{
			try
			{
				// Wait for a connection
				NamedPipeServerStream pipeServer = CreateSecurityPipeServer();
				pipeServer.BeginWaitForConnection(new AsyncCallback(WaitForConnectionCallBack), pipeServer);
			}
			catch (Exception ex)
			{
				StaticLogger.LogInfo("PipeManager : StartListen() exception : " + ex.ToString());
			}
			finally
			{
			}
		}

		internal static byte[] StructToBytes(object structObj)
		{
			int size = Marshal.SizeOf(structObj);
			IntPtr buffer = Marshal.AllocHGlobal(size);

			try
			{
				Marshal.StructureToPtr(structObj, buffer, false);
				byte[] byteData = new byte[size];
				Marshal.Copy(buffer, byteData, 0, size);
				return byteData;
			}
			catch (Exception ex)
			{
				throw ex;
			}
			finally
			{
				Marshal.FreeHGlobal(buffer);
			}
		}

		internal static TClass BytesToStruct<TClass>(byte[] byteData)
		{
			int len = byteData.Length;
			IntPtr buffer = Marshal.AllocHGlobal(len);
			try
			{
				Marshal.Copy(byteData, 0, buffer, len);
				return (TClass)Marshal.PtrToStructure(buffer, typeof(TClass));
			}
			catch (Exception ex)
			{
				throw ex;
			}
			finally
			{
				Marshal.FreeHGlobal(buffer);
			}
		}

		private NamedPipeServerStream CreateSecurityPipeServer()
		{
			PipeSecurity ps = new PipeSecurity();
			System.Security.Principal.SecurityIdentifier sid = new System.Security.Principal.SecurityIdentifier(System.Security.Principal.WellKnownSidType.WorldSid, null);
			PipeAccessRule par = new PipeAccessRule(sid, PipeAccessRights.ReadWrite, System.Security.AccessControl.AccessControlType.Allow);
			ps.AddAccessRule(par);
			StaticLogger.LogInfo("PipeManager : StartListen()");
			NamedPipeServerStream pipeServer = new NamedPipeServerStream(this.PipeName, PipeDirection.InOut, 1, PipeTransmissionMode.Byte, PipeOptions.Asynchronous, 1024, 1024, ps);

			return pipeServer;
		}

		internal void WaitForConnectionCallBack(IAsyncResult asynResult)
		{
			StaticLogger.LogInfo("PipeManager : WaitForConnectionCallBack()");
			NamedPipeServerStream pipeServer = (NamedPipeServerStream)asynResult.AsyncState;
			if (pipeServer != null)
			{
				pipeServer.EndWaitForConnection(asynResult);

				//Get the message
				TClass data = new TClass();
				byte[] buffer = new byte[Marshal.SizeOf(data)];
				try
				{
					pipeServer.Read(buffer, 0, buffer.Length);
					data = BytesToStruct<TClass>(buffer);
				}
				catch (Exception ex)
				{
					StaticLogger.LogInfo("PipeManager : WaitForConnectionCallBack() " + ex.ToString());
				}

				pipeServer.Close();
				pipeServer = null;

				pipeServer = CreateSecurityPipeServer();
				pipeServer.BeginWaitForConnection(new AsyncCallback(WaitForConnectionCallBack), pipeServer);

				if (MsgReceived != null)
				{
					StaticLogger.LogInfo("PipeManager : WaitForConnectionCallBack() callback to Main process");
					MsgReceived(data);
				}
			}
		}

		public void Send(TClass data, int timeout = 5000)
		{
			try
			{
				NamedPipeClientStream pipeClient = new NamedPipeClientStream(".", this.PipeName, PipeDirection.InOut, PipeOptions.Asynchronous);
				byte[] buffer = StructToBytes(data);
				StaticLogger.LogInfo("PipeManager : Send() buffer size : " + buffer.Length.ToString());
				pipeClient.Connect(timeout);

				pipeClient.Write(buffer, 0, buffer.Length);
			}
			catch (TimeoutException timeoutEX)
			{
				StaticLogger.LogInfo("PipeManager : Send() exception: " + timeoutEX.ToString());
			}
		}
	}

	[StructLayout(LayoutKind.Sequential, CharSet = CharSet.Auto)]
	public class ALUIPipeData
	{
		public bool? Enable;
		[MarshalAs(UnmanagedType.ByValArray, SizeConst = 8)]
		public byte[] BTMacAddress;
		[MarshalAs(UnmanagedType.ByValTStr, SizeConst = 256)]
		public string Username;
		[MarshalAs(UnmanagedType.ByValTStr, SizeConst = 256)]
		public string Password;

		public ALUIPipeData()
		{
			BTMacAddress = null;
			Username = null;
			Password = null;
			Enable = null;
		}

		public ALUIPipeData(byte[] btAddress, string username, string password)
		{
			BTMacAddress = btAddress;
			Username = username;
			Password = password;
			Enable = true;
		}

		public ALUIPipeData(byte[] btAddress)
		{
			BTMacAddress = btAddress;
			Enable = false;
		}
	}

	internal class ALUIPipeManager : PipeManager<ALUIPipeData>
	{
		public ALUIPipeManager()
		{
			this.PipeName = "ALUIPipe";
		}

		public void RecordALInfo(BluetoothAddress btAddress, string username, string password, int timeout = 5000)
		{
			StaticLogger.LogInfo("PipeManager : RecordALInfo()");
			Send(new ALUIPipeData(btAddress.ToByteArray(), username, password), timeout);
		}

		public void DeleteALInfo(BluetoothAddress btAddress, int timeout = 5000)
		{
			StaticLogger.LogInfo("PipeManager : DeleteALInfo()");
			Send(new ALUIPipeData(btAddress.ToByteArray()), timeout);
		}
	}
}
