﻿using AirFusion.Manager.DeviceManager;
using Microsoft.Win32;
using System;
using System.Drawing;
using System.IO;
using System.Net;
using System.Net.NetworkInformation;
using System.Text;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Imaging;

namespace AirFusion
{
	public static class Utils
	{
		#region Related to DPI

		/// <summary>
		/// This is the parameter for GetDeviceCaps() API of Win32
		/// <remarks>
		/// http://pinvoke.net/default.aspx/gdi32/GetDeviceCaps.html
		/// http://msdn.microsoft.com/en-us/library/windows/desktop/dd144877(v=vs.85).aspx
		/// </remarks>
		/// </summary>
		private const int _LogPixelsX = 88;

		/// <summary>
		/// Gets current DPI in the machine. The method is based on other In-house application project.
		/// </summary>
		/// <returns>
		/// The value of current DPI
		/// </returns>
		public static double GetDPI()
		{
			try
			{
				IntPtr h = Win32Api.GetDC(IntPtr.Zero);
				double dpiX = Win32Api.GetDeviceCaps(h, _LogPixelsX);
				if (dpiX <= 96.0)
				{
					return 96;
				}
				else if (dpiX <= 120.0 && dpiX > 96.0)
				{
					return 120.0;
				}
				else if (dpiX <= 144.0 && dpiX > 120.0)
				{
					return 144.0;
				}
				else if (dpiX > 144.0)
				{
					return 192.0;
				}
			}
			catch
			{
			}

			return 96;
		}

		/// <summary>
		/// Gets current DPI ratio with DPI 96
		/// </summary>
		/// <returns>
		/// The value of current DPI ratio
		/// </returns>
		public static double GetDPIRatio(bool isNeedAdjust = true)
		{
			try
			{
				IntPtr h = Win32Api.GetDC(IntPtr.Zero);
				double dpiX = Win32Api.GetDeviceCaps(h, _LogPixelsX);

				if (isNeedAdjust)
				{
					if (dpiX <= 96.0)
						return 1.0;
					else if (dpiX <= 120.0 && dpiX > 96.0)
						return 1.25;
					else if (dpiX <= 144.0 && dpiX > 120.0)
						return 1.5;
					else if (dpiX <= 192.0 && dpiX > 144.0)
						return 2.0;
				}
				else
					return dpiX / 96.0;
			}
			catch
			{
			}

			return 1.0;
		}

		#endregion Related to DPI

		#region Related to OS Account

		/// <summary>
		/// Gets the account display name of OS
		/// </summary>
		/// <returns>
		/// The current account display name
		/// </returns>
		public static string GetAccountDisplayName()
		{
			string displayName = Environment.UserName;
			StringBuilder buffer = new StringBuilder(1024);
			int size = buffer.Capacity;
			if (Win32Api.GetUserNameEx((int)EXTENDED_NAME_FORMAT.NameDisplay, buffer, ref size) != 0)
			{
				if (buffer.Length != 0)
				{
					displayName = buffer.ToString();
				}
			}

			return displayName;
		}

		/// <summary>
		/// Gets the default account name of OS
		/// </summary>
		/// <returns>
		/// The current default account name
		/// </returns>
		public static string GetAccountDefaultName()
		{
			string cpAccountName = string.Empty;
			bool findLiveId = false;
			RegistryKey reg = Registry.CurrentUser.OpenSubKey(@"Software\Microsoft\IdentityCRL\UserExtendedProperties", false);
			if (reg != null)
			{
				string[] subkeys = reg.GetSubKeyNames();
				if (subkeys != null && subkeys.Length >= 1)
				{
					//Find live-id account e-mail
					cpAccountName = subkeys[0];
					findLiveId = true;
				}

				reg.Close();
				reg = null;
			}

			if (!findLiveId)
			{
				cpAccountName = Environment.UserName;
			}

			return cpAccountName;
		}

		#endregion Related to OS Account

		#region Related to Image

		/// <summary>
		/// Creates the bitmap source from file path
		/// </summary>
		/// <param name="iconFile">
		/// The file path of image
		/// </param>
		/// <returns>
		/// The bitmap
		/// </returns>
		public static BitmapSource CreateBitmapSourceByFile(string iconFile)
		{
			//Bitmap bitmapIcon = new Bitmap(System.Drawing.Image.FromFile(iconFile));
			//System.Windows.Media.Imaging.BitmapSource bitmapSource = System.Windows.Interop.Imaging.CreateBitmapSourceFromHBitmap(
			//			bitmapIcon.GetHbitmap(),
			//			IntPtr.Zero,
			//			System.Windows.Int32Rect.Empty,
			//			System.Windows.Media.Imaging.BitmapSizeOptions.FromWidthAndHeight(bitmapIcon.Width, bitmapIcon.Height));

			//bitmapSource.Freeze();

			BitmapImage bitmapImage = new BitmapImage();
			using (FileStream fstream = File.OpenRead(iconFile))
			{
				bitmapImage.BeginInit();
				bitmapImage.CacheOption = BitmapCacheOption.OnLoad;
				bitmapImage.StreamSource = fstream;
				bitmapImage.EndInit();
			}

			BitmapSource bitmapSource = bitmapImage;

			return bitmapSource;
		}

		/// <summary>
		/// Creates the bitmap from bitmap source
		/// </summary>
		/// <param name="bitmapsource">
		/// The bitmap source
		/// </param>
		/// <returns>
		/// The bitmap
		/// </returns>
		public static System.Drawing.Bitmap CreateBitmapFromSource(BitmapSource bitmapsource)
		{
			System.Drawing.Bitmap bitmap;

			using (MemoryStream outStream = new MemoryStream())
			{
				BitmapEncoder enc = new BmpBitmapEncoder();
				enc.Frames.Add(BitmapFrame.Create(bitmapsource));
				enc.Save(outStream);
				bitmap = new System.Drawing.Bitmap(outStream);
			}

			return bitmap;
		}

		/// <summary>
		/// Saves framework element as image to file
		/// </summary>
		/// <param name="visual">
		/// The target UI framework element
		/// </param>
		/// <param name="fileFolder">
		/// The folder of output file
		/// </param>
		/// <returns></returns>
		public static string SaveFrameworkElementToImageFile(FrameworkElement visual, string fileFolder)
		{
			try
			{
				if (!Directory.Exists(fileFolder))
				{
					Directory.CreateDirectory(fileFolder);
				}

				LogtoFile.WriteLog("", Configs.DebugLevelValue, string.Format("FrameworkElement Width: {0}, Height: {1}", (int)visual.ActualWidth, (int)visual.ActualHeight), "");

				PngBitmapEncoder encoder = new PngBitmapEncoder();
				RenderTargetBitmap bitmap = new RenderTargetBitmap((int)visual.ActualWidth, (int)visual.ActualHeight, 96, 96, PixelFormats.Pbgra32);
				bitmap.Render(visual);
				bitmap.Freeze();

				BitmapFrame frame = BitmapFrame.Create(bitmap);
				encoder.Frames.Add(frame);

				string fileName = Path.Combine(fileFolder, string.Format("{0}.png", DateTime.Now.ToString("yyyyMMddHHmmssfff")));
				using (var stream = File.Create(fileName))
				{
					encoder.Save(stream);
				}

				return fileName;
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
				return string.Empty;
			}
		}

		#endregion Related to Image

		#region Related to Device Information

		/// <summary>
		/// Mapps the rotation enumeration data to angle
		/// </summary>
		/// <param name="rotation">
		/// The rotation data enumeration
		/// </param>
		/// <returns>
		/// The angle value
		/// </returns>
		public static double MappingRotationToAngle(AFUserDevice.DisplayInfo.DisplayRotation rotation)
		{
			double angle = 0;

			try
			{
				var type = typeof(AFUserDevice.DisplayInfo.DisplayRotation);
				var memInfo = type.GetMember(rotation.ToString());
				var attributes = memInfo[0].GetCustomAttributes(typeof(AirFusion.Common.ParameterValue), false);
				var finalangle = ((AirFusion.Common.ParameterValue)attributes[0]).ToString();

				angle = (double)Convert.ToInt32(finalangle);
			}
			catch
			{
			}

			return angle;
		}

		/// <summary>
		/// Detects the system permission of phone from AFUserDevice class
		/// </summary>
		public static bool DetectPhoneHasSystemPermission(AFUserDevice device)
		{
			if (null != device)
			{
				if (device.HasSystemPermission.HasValue)
				{
					return device.HasSystemPermission.Value;
				}

				LogtoFile.WriteLog("", Configs.DebugLevelValue, "device.HasSystemPermission has not value", "");
			}
			else
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "device is null", "");
			}

			// Set the default value of HasSystemPermission is false
			return false;
		}

		#endregion Related to Device Information

		#region Related to folder

		/// <summary>
		/// Gets download folder for this PC
		/// </summary>
		public static string GetDownloadFolder()
		{
			string downloads;

			Win32Api.SHGetKnownFolderPath(Win32Api.Downloads, 0, IntPtr.Zero, out downloads);

			return downloads;
		}

		#endregion Related to folder

		#region Google Analytics

		/// <summary>
		/// sned userbehavior to google analytics
		/// </summary>
		private static string s_utmhn = "acerextend2.com"; //GA設定追蹤的首頁(需與ga系統上設定吻合)

		private static string s_utmac = "UA-64451741-2"; //追蹤單號(需與ga系統上設定吻合)
		private static string s_utmwv = "5.2.9";     // GA Version 5.2.9
		private static string s_utms = "1";          //
		private static string s_utmul = "zh-tw";           // language
		private static string s_utmr = "0";          // refer, ~=none
		private static string s_utmp = "/af/datacollect/";          // URI
		private static string c_domain_hash = "84097795";  // Domain hash
		private static int c_session = 1;    // record by client and ++?
		private static int c_source = 1;     // record by client and ++?

		public static void GASend(string set_KeyName, string set_KeyValue, string DeviceID)
		{
			//first. check network available
			bool isNetworkAvailable = NetworkInterface.GetIsNetworkAvailable();
			if (!isNetworkAvailable)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Google Analytics must be connected to network, but now do not connected it.", "");
			}
			else
			{
				//second. send httprequest to google analytics.
				try
				{
					StringBuilder builder = new StringBuilder();
					builder.Append("http://www.google-analytics.com/__utm.gif");
					StringBuilder cookie = new StringBuilder();
					builder.Append("?utmwv=" + s_utmwv);
					builder.Append("&utms=" + s_utms);
					Random s_utmn = new Random(Guid.NewGuid().GetHashCode());
					builder.Append("&utmn=" + s_utmn.Next(0, 2000000000));
					builder.Append("&utmhn=" + s_utmhn);

					//setting user-define
					string str_utme = "&utme=8(1!" + set_KeyName + "*5!" + set_KeyName + "_DeviceID" + ")9(1!" + set_KeyValue + "*5!" + DeviceID + "/" + set_KeyValue + ")";
					builder.Append(str_utme);
					s_utmul = System.Globalization.CultureInfo.CurrentCulture.ToString();
					builder.Append("&utmul=" + s_utmul);
					builder.Append("&utmr=" + s_utmr);
					builder.Append("&utmp=" + s_utmp);
					builder.Append("&utmac=" + s_utmac);

					// utam
					c_domain_hash = domainhash();
					cookie.Append("&utmcc=__utma=" + c_domain_hash);
					Random c_uid = new Random(Guid.NewGuid().GetHashCode());
					cookie.Append("." + c_uid.Next(0, 2000000000));
					IFormatProvider culture = new System.Globalization.CultureInfo(s_utmul);
					DateTime InitialTime = DateTime.ParseExact("20150701000000", "yyyyMMddHHmmss", culture); //GA正式統計起始日期
					DateTime CurrentTime = DateTime.Now;
					TimeSpan ts0 = new TimeSpan(InitialTime.Ticks);
					TimeSpan ts3 = new TimeSpan(CurrentTime.Ticks);
					TimeSpan Timestamp = ts3.Subtract(ts0).Duration();
					cookie.Append("." + (int)Timestamp.TotalSeconds); // First
					cookie.Append("." + (int)Timestamp.TotalSeconds); // Last
					cookie.Append("." + (int)Timestamp.TotalSeconds); // Current
					cookie.Append("." + c_session);

					// utmz
					cookie.Append(";+__utmz=" + c_domain_hash);

					//get UIVersion
					string UIVersion = String.Empty;
					try
					{
						bool isX86;
						object oRK = RegistryUtility.GetRegistry(RegistryHive.LocalMachine, "SOFTWARE\\Wow6432Node\\OEM\\AcerEXTEND\\Installer Property", "MSI_ProductDisplayVersion", RegistryView.Registry64, out isX86);
						UIVersion = (string)oRK;
					}
					catch
					{
						UIVersion = "unknow UIVersion";
					}

					//透過客戶開發-來源與媒介得知使用者的UI版號
					cookie.Append("." + (int)Timestamp.TotalSeconds); // First
					cookie.Append("." + c_session);
					cookie.Append("." + c_source);
					cookie.Append(".utmcsr=" + DeviceID + "|utmccn=" + DeviceID + "|utmcmd=" + UIVersion);

					// utmv
					cookie.Append(";+__utmv=" + c_domain_hash);
					cookie.Append("." + DeviceID);
					cookie.Append(";&utmu=q~");
					builder.Append(string.Format(cookie.ToString()));

					try
					{
						HttpWebRequest GArequest = (HttpWebRequest)WebRequest.Create(string.Format(builder.ToString()));
						GArequest.Timeout = 5000;
						GArequest.Method = "GET";
						GArequest.BeginGetResponse(new AsyncCallback(RespCallback), GArequest);
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "Google analytics send (" + set_KeyName + "=>" + set_KeyValue + " from " + DeviceID + ") Success.", "");
					}
					catch (Exception ex)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "Google analytics http request Failure.", "");
					}
				}
				catch (Exception ex)
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Google analytics GAsend() Failure.", "");
				}
			}
		}

		private static void RespCallback(IAsyncResult asynchronousResult)
		{
			HttpWebRequest request = null;
			HttpWebResponse response = null;

			try
			{
				request = (HttpWebRequest)asynchronousResult.AsyncState;
				response = (HttpWebResponse)request.EndGetResponse(asynchronousResult);
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Google Analytics RespCallback() Failure.", "");
			}
			finally
			{
				if (response != null)
					response.Close(); // Release the HttpWebResponse
			}
		}

		public static string domainhash()
		{
			string buffer;
			int shifter;
			int a = 0;
			int c = 0;
			for (int i = s_utmhn.Length - 1; i >= 0; i--)
			{
				buffer = "" + charCodeAt(i, s_utmhn);
				shifter = Int32.Parse(buffer);
				a = ((a * (int)Math.Pow(2, 6)) & 268435455) + shifter + (shifter * (int)Math.Pow(2, 14));
				c = a & 266338304;
				if (c != 0)
					a = a ^ c / (int)Math.Pow(2, 21);
			}
			return a.ToString();
		}

		public static byte charCodeAt(int n, string str)
		{
			byte[] bytes = System.Text.Encoding.UTF8.GetBytes(str);
			return bytes[n];
		}

		#endregion Google Analytics
	}
}