﻿using System;
using System.ComponentModel;

namespace AirFusion.Common.Converters
{
	/// <summary>
	/// Provides a type converter to convert System.DateTime objects to and from various other representations.
	/// </summary>
	public class WmiDateTimeConverter : TypeConverter
	{
		/// <summary>
		/// Returns whether this converter can convert an object of the given type to the type of this converter, using the specified context.
		/// </summary>
		/// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext" /> that provides a format context.</param>
		/// <param name="sourceType">A <see cref="T:System.Type" /> that represents the type you want to convert from.</param>
		/// <returns>
		/// true if this converter can perform the conversion; otherwise, false.
		/// </returns>
		public override bool CanConvertFrom(ITypeDescriptorContext context, Type sourceType)
		{
			return sourceType.Equals(typeof(DateTime)) || sourceType.Equals(typeof(string));
		}

		/// <summary>
		/// Returns whether this converter can convert the object to the specified type, using the specified context.
		/// </summary>
		/// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext" /> that provides a format context.</param>
		/// <param name="destinationType">A <see cref="T:System.Type" /> that represents the type you want to convert to.</param>
		/// <returns>
		/// true if this converter can perform the conversion; otherwise, false.
		/// </returns>
		public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType)
		{
			return destinationType.Equals(typeof(string));
		}

		/// <summary>
		/// Returns whether changing a value on this object requires a call to <see cref="M:System.ComponentModel.TypeConverter.CreateInstance(System.Collections.IDictionary)" /> to create a new value, using the specified context.
		/// </summary>
		/// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext" /> that provides a format context.</param>
		/// <returns>
		/// true if changing a property on this object requires a call to <see cref="M:System.ComponentModel.TypeConverter.CreateInstance(System.Collections.IDictionary)" /> to create a new value; otherwise, false.
		/// </returns>
		public override bool GetCreateInstanceSupported(ITypeDescriptorContext context)
		{
			return false;
		}

		/// <summary>
		/// Converts the given object to the type of this converter, using the specified context and culture information.
		/// </summary>
		/// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext" /> that provides a format context.</param>
		/// <param name="culture">The <see cref="T:System.Globalization.CultureInfo" /> to use as the current culture.</param>
		/// <param name="value">The <see cref="T:System.Object" /> to convert.</param>
		/// <returns>
		/// An <see cref="T:System.Object" /> that represents the converted value.
		/// </returns>
		/// <exception cref="System.ArgumentOutOfRangeException" />
		/// <exception cref="System.NotSupportedException" />
		public override object ConvertFrom(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value)
		{
			if (value == null)
			{
				return null;
			}
			else if (value is DateTime)
			{
				return value;
			}
			else if (value is string)
			{
				DateTime initializer = DateTime.MinValue;
				int year = initializer.Year;
				int month = initializer.Month;
				int day = initializer.Day;
				int hour = initializer.Hour;
				int minute = initializer.Minute;
				int second = initializer.Second;
				long ticks = 0;
				string dmtf = value as string;
				DateTime datetime = DateTime.MinValue;
				string tempString = string.Empty;
				if ((dmtf == null))
				{
					throw new System.ArgumentOutOfRangeException();
				}
				if ((dmtf.Length == 0))
				{
					throw new System.ArgumentOutOfRangeException();
				}
				if ((dmtf.Length != 25))
				{
					throw new System.ArgumentOutOfRangeException();
				}
				try
				{
					tempString = dmtf.Substring(0, 4);
					if (("****" != tempString))
					{
						year = int.Parse(tempString);
					}
					tempString = dmtf.Substring(4, 2);
					if (("**" != tempString))
					{
						month = int.Parse(tempString);
					}
					tempString = dmtf.Substring(6, 2);
					if (("**" != tempString))
					{
						day = int.Parse(tempString);
					}
					tempString = dmtf.Substring(8, 2);
					if (("**" != tempString))
					{
						hour = int.Parse(tempString);
					}
					tempString = dmtf.Substring(10, 2);
					if (("**" != tempString))
					{
						minute = int.Parse(tempString);
					}
					tempString = dmtf.Substring(12, 2);
					if (("**" != tempString))
					{
						second = int.Parse(tempString);
					}
					tempString = dmtf.Substring(15, 6);
					if (("******" != tempString))
					{
						ticks = (long.Parse(tempString) * ((long)((System.TimeSpan.TicksPerMillisecond / 1000))));
					}
					if (((((((((year < 0)
								|| (month < 0))
								|| (day < 0))
								|| (hour < 0))
								|| (minute < 0))
								|| (minute < 0))
								|| (second < 0))
								|| (ticks < 0)))
					{
						throw new System.ArgumentOutOfRangeException();
					}
				}
				catch (System.Exception e)
				{
					throw new System.ArgumentOutOfRangeException(null, e.Message);
				}
				datetime = new DateTime(year, month, day, hour, minute, second, 0);
				datetime = datetime.AddTicks(ticks);
				TimeSpan tickOffset = System.TimeZone.CurrentTimeZone.GetUtcOffset(datetime);
				int UTCOffset = 0;
				int OffsetToBeAdjusted = 0;
				long OffsetMins = ((long)((tickOffset.Ticks / System.TimeSpan.TicksPerMinute)));
				tempString = dmtf.Substring(22, 3);
				if ((tempString != "******"))
				{
					tempString = dmtf.Substring(21, 4);
					try
					{
						UTCOffset = int.Parse(tempString);
					}
					catch (System.Exception e)
					{
						throw new System.ArgumentOutOfRangeException(null, e.Message);
					}
					OffsetToBeAdjusted = ((int)((OffsetMins - UTCOffset)));
					datetime = datetime.AddMinutes(((double)(OffsetToBeAdjusted)));
				}
				return datetime;
			}
			else
			{
				throw new NotSupportedException("The type '" + value.GetType() + "' is not supported.");
			}
		}

		/// <summary>
		/// Converts the given value object to the specified type, using the specified context and culture information.
		/// </summary>
		/// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext" /> that provides a format context.</param>
		/// <param name="culture">A <see cref="T:System.Globalization.CultureInfo" />. If null is passed, the current culture is assumed.</param>
		/// <param name="value">The <see cref="T:System.Object" /> to convert.</param>
		/// <param name="destinationType">The <see cref="T:System.Type" /> to convert the <paramref name="value" /> parameter to.</param>
		/// <returns>
		/// An <see cref="T:System.Object" /> that represents the converted value.
		/// </returns>
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.NotSupportedException" />
		public override object ConvertTo(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value, Type destinationType)
		{
			if (!(value is DateTime))
			{
				throw new ArgumentException("The value you specified is not a DateTime value.");
			}

			if (destinationType == null)
			{
				throw new ArgumentNullException("destinationType");
			}
			if (destinationType.Equals(typeof(DateTime)))
			{
				return value;
			}
			else if (destinationType.Equals(typeof(string)))
			{
				DateTime date = (DateTime)value;
				string utcString = string.Empty;
				System.TimeSpan tickOffset = System.TimeZone.CurrentTimeZone.GetUtcOffset(date);
				long OffsetMins = ((long)((tickOffset.Ticks / System.TimeSpan.TicksPerMinute)));
				if ((System.Math.Abs(OffsetMins) > 999))
				{
					date = date.ToUniversalTime();
					utcString = "+000";
				}
				else
				{
					if ((tickOffset.Ticks >= 0))
					{
						utcString = string.Concat("+", ((long)((tickOffset.Ticks / System.TimeSpan.TicksPerMinute))).ToString().PadLeft(3, '0'));
					}
					else
					{
						string strTemp = ((long)(OffsetMins)).ToString();
						utcString = string.Concat("-", strTemp.Substring(1, (strTemp.Length - 1)).PadLeft(3, '0'));
					}
				}
				string dmtfDateTime = ((int)(date.Year)).ToString().PadLeft(4, '0');
				dmtfDateTime = string.Concat(dmtfDateTime, ((int)(date.Month)).ToString().PadLeft(2, '0'));
				dmtfDateTime = string.Concat(dmtfDateTime, ((int)(date.Day)).ToString().PadLeft(2, '0'));
				dmtfDateTime = string.Concat(dmtfDateTime, ((int)(date.Hour)).ToString().PadLeft(2, '0'));
				dmtfDateTime = string.Concat(dmtfDateTime, ((int)(date.Minute)).ToString().PadLeft(2, '0'));
				dmtfDateTime = string.Concat(dmtfDateTime, ((int)(date.Second)).ToString().PadLeft(2, '0'));
				dmtfDateTime = string.Concat(dmtfDateTime, ".");
				System.DateTime dtTemp = new System.DateTime(date.Year, date.Month, date.Day, date.Hour, date.Minute, date.Second, 0);
				long microsec = ((long)((((date.Ticks - dtTemp.Ticks) * 1000) / System.TimeSpan.TicksPerMillisecond)));
				string strMicrosec = ((long)(microsec)).ToString();
				if ((strMicrosec.Length > 6))
				{
					strMicrosec = strMicrosec.Substring(0, 6);
				}
				dmtfDateTime = string.Concat(dmtfDateTime, strMicrosec.PadLeft(6, '0'));
				dmtfDateTime = string.Concat(dmtfDateTime, utcString);
				return dmtfDateTime;
			}
			else
			{
				throw new NotSupportedException("The destination type '" + destinationType.Name + "' is not supported.");
			}
		}
	}
}
