﻿//   Class: Logger
// Purpose: This is the class with static function to log the file.
//  Author: Seoker Wang 2012-04 modified from Bryant Huang (C++ MFC version)
// Change Logs:
//      Date        Name           Description
//      2013-04-18  Seoker Wang    Modify to make more stable. (Support only filename as filePath)
//      2013-05-24  Seoker Wang    Fixes Uninitialize() throws exceptions.
//      2014-12-15  Seoker Wang    Fixes race conditions.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text.RegularExpressions;

namespace AirFusion.Common
{
	#region Enumeration

	/// <summary>
	/// Specifies the log level.
	/// </summary>
	public enum LogLevel
	{
		/// <summary>
		/// Trace.
		/// </summary>
		Trace = 0,

		/// <summary>
		/// Debug.
		/// </summary>
		Debug = 1,

		/// <summary>
		/// Information.
		/// </summary>
		Info = 2,

		/// <summary>
		/// Warning.
		/// </summary>
		Warn = 3,

		/// <summary>
		/// Error.
		/// </summary>
		Error = 4,

		/// <summary>
		/// Fatal.
		/// </summary>
		Fatal = 5,

		/// <summary>
		/// Off.
		/// </summary>
		Off = 6
	}

	/// <summary>
	/// Specifies the log type.
	/// </summary>
	public enum LogType
	{
		/// <summary>
		/// Writes into single log file by day.
		/// </summary>
		LogDaily,

		/// <summary>
		/// Writes into single log file by file size.
		/// </summary>
		LogFileSize,

		/// <summary>
		/// Writes into log files by day.
		/// </summary>
		LogDailyRolling,

		/// <summary>
		/// Writes into log files by file size.
		/// </summary>
		LogFileSizeRolling
	}

	/// <summary>
	/// Specifies the return code.
	/// </summary>
	public enum LogReturnCode
	{
		/// <summary>
		/// Indicates the operation is succussful.
		/// </summary>
		Success,

		/// <summary>
		/// Indicates there's an IO error during the operation.
		/// </summary>
		IOError,

		/// <summary>
		/// Indicates the specified file does not exist.
		/// </summary>
		FileNotExist,

		/// <summary>
		/// Indicates the component is not initialized yet.
		/// </summary>
		NotInitialized,

		/// <summary>
		/// Indicates the parameter is incorrect.
		/// </summary>
		ParameterError,

		/// <summary>
		/// Indicates there's an unknown error during the operation.
		/// </summary>
		UnknownError
	}

	#endregion Enumeration

	#region EpdaLogger

	/// <summary>
	/// Provides the methods of logger.
	/// </summary>
	public static class StaticLogger
	{
		#region Constant

		private const ulong _maxLogLength = 10240; // 10K in a single message

		#endregion Constant

		#region Static Variables

		private static ulong _fileCount = 10;
		private static ulong _fileMaxSize = 1024 * 1024;
		private static bool _isInitialized = false;
		private static bool _recordMethodName = true;

		private static string _filePath = "";
		private static string _logFolder = null;
		private static string _fileNameWithoutExtension = null;
		private static string _fileExtension = null; // include dot

		private static StreamWriter _fileWriterObject = null;
		private static Object _operationLocker = new Object();
		private static Object _fileWriterObjectLocker = new Object();
		private static List<string> _logFileList = null;
		private static DateTime _lastLogDate = DateTime.Now;

		#endregion Static Variables

		#region Property

		/// <summary>
		/// Gets the log level.
		/// </summary>
		public static LogLevel LogLevel { get; private set; }

		/// <summary>
		/// Gets the type of the logger.
		/// </summary>
		public static LogType LogType { get; private set; }

		/// <summary>
		/// Gets the file path of the log file.
		/// </summary>
		public static string FilePath
		{
			get
			{
				return string.IsNullOrEmpty(_filePath) ? string.Empty : _filePath;
			}
		}

		/// <summary>
		/// The error message generated from the last error occurred.
		/// </summary>
		public static string LastErrorMessage { get; private set; }

		#endregion Property

		#region Static Method

		static StaticLogger()
		{
			LogLevel = LogLevel.Debug;
			LogType = LogType.LogFileSize;
			_logFileList = new List<string>();
		}

		/// <summary>
		/// Initializes the logger component.
		/// </summary>
		/// <param name="filePath">The file path of log files.</param>
		/// <param name="logLevel">The log level.</param>
		/// <param name="logType">The rolling type.</param>
		/// <param name="appendDateTime">true if to append date/time.</param>
		/// <param name="recordMethodName">true if to record method name.</param>
		/// <returns></returns>
		[DebuggerStepThrough]
		public static LogReturnCode Initialize(string filePath, LogLevel logLevel = LogLevel.Debug, LogType logType = LogType.LogFileSize, bool appendDateTime = true, bool recordMethodName = true)
		{
			LastErrorMessage = "";
			LogReturnCode ret;

			// check parameters
			if (string.IsNullOrWhiteSpace(filePath))
			{
				LastErrorMessage = "Please specify log file path.";
				return LogReturnCode.ParameterError;
			}

			// check prerequisite
			// check if it's already initialized, if yes, uninitialize it first
			if ((ret = UnInitialize()) != LogReturnCode.Success)
			{
				LastErrorMessage = "Uninitialization failed.";
				return ret;
			}

			Debug.Assert(_operationLocker != null);
			lock (_operationLocker)
			{
				// set properties
				LogLevel = logLevel;
				LogType = logType;

				// set variables
				_fileWriterObject = null;
				_filePath = Path.GetFullPath(filePath);
				_recordMethodName = recordMethodName;

				// parse filepath
				try
				{
					_logFolder = Path.GetDirectoryName(_filePath);
					_fileNameWithoutExtension = Path.GetFileNameWithoutExtension(_filePath);
					_fileExtension = Path.GetExtension(_filePath);
				}
				catch (Exception ex)
				{
					LastErrorMessage = ex.Message + "\r\nFailed getting file path.";
					return LogReturnCode.ParameterError;
				}

				// debug assertion
				Debug.Assert(_logFolder != null);
				Debug.Assert(_fileNameWithoutExtension != null);
				Debug.Assert(_fileExtension != null);

				// if log level is Off, do nothing
				if (LogLevel == LogLevel.Off)
				{
					return LogReturnCode.Success;
				}

				// find current log files in the folder
				Debug.Assert(_logFileList != null);
				lock (_logFileList)
				{
					_logFileList.Clear();

					try
					{
						if (!Directory.Exists(_logFolder))
						{
							Directory.CreateDirectory(_logFolder);
						}
					}
					catch (Exception ex)
					{
						LastErrorMessage = ex.Message + "\r\nFailed creating folder.";
						return LogReturnCode.IOError;
					}

					try
					{
						if (!Directory.Exists(_logFolder))
						{
							Directory.CreateDirectory(_logFolder);
						}

						foreach (var logFilePath in Directory.EnumerateFiles(_logFolder, _fileNameWithoutExtension + "*" + _fileExtension))
						{
							_logFileList.Add(logFilePath);
						}
					}
					catch (Exception ex)
					{
						LastErrorMessage = ex.Message + "\r\nFailed enumerating files.";
						return LogReturnCode.IOError;
					}
				}

				// set initialize flag
				_isInitialized = true;
			}

			// set new file path
			if ((ret = SetNewFilePath(appendDateTime)) != LogReturnCode.Success)
			{
				LastErrorMessage = LastErrorMessage + "\r\nFailed setting new path.";
				return ret;
			}

			// remove other log files
			RemoveLog();

			return LogReturnCode.Success;
		}

		/// <summary>
		/// Uninitializes the logger component.
		/// </summary>
		/// <returns></returns>
		[DebuggerStepThrough]
		public static LogReturnCode UnInitialize()
		{
			LastErrorMessage = "";

			Debug.Assert(_operationLocker != null);
			lock (_operationLocker)
			{
				if (_isInitialized == false)
				{
					return LogReturnCode.Success;
				}

				// close the log file
				try
				{
					lock (_fileWriterObjectLocker)
					{
						if (_fileWriterObject != null)
						{
							_fileWriterObject.Flush();
							_fileWriterObject.Close();
							_fileWriterObject = null;
						}
					}
				}
				catch (Exception ex)
				{
					LastErrorMessage = ex.Message;
					return LogReturnCode.IOError;
				}

				_isInitialized = false;

				return LogReturnCode.Success;
			}
		}

		/// <summary>
		/// Sets the log parameter such as file count or file size.
		/// </summary>
		/// <param name="fileCount">The file count</param>
		/// <param name="fileMaxSize">The file maximum size</param>
		/// <returns></returns>
		[DebuggerStepThrough]
		public static LogReturnCode SetParameter(ulong fileCount, ulong fileMaxSize)
		{
			LastErrorMessage = string.Empty;

			Debug.Assert(_operationLocker != null);
			lock (_operationLocker)
			{
				_fileCount = fileCount;
				_fileMaxSize = fileMaxSize;
			}

			return LogReturnCode.Success;
		}

		/// <summary>
		/// Logs a message.
		/// </summary>
		/// <param name="logLevel">The log level.</param>
		/// <param name="functionName">The function or component name.</param>
		/// <param name="message">The log message.</param>
		/// <returns></returns>
		[DebuggerStepThrough]
		public static LogReturnCode Log(LogLevel logLevel, string functionName, string message)
		{
			LastErrorMessage = string.Empty;

			try
			{
				Debug.Assert(_operationLocker != null);
				lock (_operationLocker)
				{
					// check prerequsite
					if (_isInitialized == false)
					{
						return LogReturnCode.NotInitialized;
					}

					// roll the log files
					LogReturnCode ret = RollingFile();
					if (ret != LogReturnCode.Success)
					{
						return ret;
					}

					Debug.Assert(_fileWriterObjectLocker != null);
					lock (_fileWriterObjectLocker)
					{
						// check prerequisite
						if (_fileWriterObject == null)
						{
							return LogReturnCode.FileNotExist;
						}
						if (LogLevel == LogLevel.Off || logLevel < LogLevel)
						{
							return LogReturnCode.Success;
						}

						message = message ?? string.Empty;

						// format the message
						string logMessage = DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss.fff\t[") +
							System.Threading.Thread.CurrentThread.ManagedThreadId.ToString("X4") + "]\t<" +
							logLevel.ToString() + ">\t";

						logMessage += (string.IsNullOrEmpty(functionName)) ? message : "[" + functionName + "]\t" + message;

						// if the length is too long, trim the right substring
						if (logMessage.Length > (int)_maxLogLength)
						{
							logMessage = logMessage.Substring(0, (int)_maxLogLength);
						}

						// write the log
						_fileWriterObject.WriteLine(logMessage);

						// set last log date
						_lastLogDate = DateTime.Now;

						// also output to the debug channel
						Debug.WriteLine("[" + logLevel.ToString() + "]" + logMessage);

						return LogReturnCode.Success;
					}
				}
			}
			catch
			{
				return LogReturnCode.UnknownError;
			}
		}

		/// <summary>
		/// Logs a message.
		/// </summary>
		/// <param name="logLevel">The log level.</param>
		/// <param name="message">The log text.</param>
		/// <returns></returns>
		[DebuggerStepThrough]
		public static LogReturnCode Log(LogLevel logLevel, string message)
		{
			string methodName = null;
			if (_recordMethodName)
			{
				// Get the class type and method
				methodName = "UnknownMethod";
				StackTrace stackTrace = new StackTrace(false);
				StackFrame stackFrame = null;
				MethodBase method = null;
				for (int index = 0; index < stackTrace.FrameCount; ++index)
				{
					stackFrame = stackTrace.GetFrame(index);
					method = stackFrame.GetMethod();

					if (method.DeclaringType != null && method.DeclaringType.Name.CompareTo(typeof(StaticLogger).Name) != 0)
					{
						break;
					}
					stackFrame = null;
					method = null;
				}
				if (stackFrame != null)
				{
					methodName = (method.DeclaringType != null ? GetFullName(method.DeclaringType) + method.DeclaringType.Name + "." : "") + method.Name + "()";
				}
			}

			return Log(logLevel, methodName, message);
		}

		/// <summary>
		/// Logs with trace log level.
		/// </summary>
		/// <param name="message">The log message.</param>
		/// <returns></returns>
		[DebuggerStepThrough]
		public static LogReturnCode LogTrace(string message)
		{
			return Log(LogLevel.Trace, message);
		}

		/// <summary>
		/// Logs with debug log level.
		/// </summary>
		/// <param name="message">The log message.</param>
		/// <returns></returns>
		[DebuggerStepThrough]
		public static LogReturnCode LogDebug(string message)
		{
			return Log(LogLevel.Debug, message);
		}

		/// <summary>
		/// Logs with information log level.
		/// </summary>
		/// <param name="message">The log message.</param>
		/// <returns></returns>
		[DebuggerStepThrough]
		public static LogReturnCode LogInfo(string message)
		{
			return Log(LogLevel.Info, message);
		}

		/// <summary>
		/// Logs with warning log level.
		/// </summary>
		/// <param name="message">The log message.</param>
		/// <returns></returns>
		[DebuggerStepThrough]
		public static LogReturnCode LogWarn(string message)
		{
			return Log(LogLevel.Warn, message);
		}

		/// <summary>
		/// Logs with error log level.
		/// </summary>
		/// <param name="message">The log message.</param>
		/// <returns></returns>
		[DebuggerStepThrough]
		public static LogReturnCode LogError(string message)
		{
			return Log(LogLevel.Error, message);
		}

		/// <summary>
		/// Logs with fatal log level.
		/// </summary>
		/// <param name="message">The log message.</param>
		/// <returns></returns>
		[DebuggerStepThrough]
		public static LogReturnCode LogFatal(string message)
		{
			return Log(LogLevel.Fatal, message);
		}

		/// <summary>
		/// Tracks the activity of entering a function.
		/// </summary>
		/// <returns></returns>
		[DebuggerStepThrough]
		public static LogReturnCode TrackEnter(string methodName = null)
		{
			StackTrace trace = new StackTrace(false);
			if (methodName == null)
			{
				StackFrame frame = trace.GetFrame(1);
				MethodBase method = (frame != null) ? frame.GetMethod() : null;
				methodName = method != null ? (method.DeclaringType != null ? GetFullName(method.DeclaringType) + "." : "") + method.Name : "unknown method";
			}

			return Log(LogLevel.Trace, ">>>>>> ========= Enter [" + methodName + "] ========= >>>>>>");
		}

		/// <summary>
		/// Tracks the activity of leaving a function.
		/// </summary>
		/// <returns></returns>
		[DebuggerStepThrough]
		public static LogReturnCode TrackLeave(string methodName = null)
		{
			StackTrace trace = new StackTrace(false);
			if (methodName == null)
			{
				StackFrame frame = trace.GetFrame(1);
				MethodBase method = (frame != null) ? frame.GetMethod() : null;
				methodName = method != null ? (method.DeclaringType != null ? GetFullName(method.DeclaringType) + "." : "") + method.Name : "unknown method";
			}

			return Log(LogLevel.Trace, "<<<<<< ========= Leave [" + methodName + "] ========= <<<<<<");
		}

		/// <summary>
		/// Removes the rolling log.
		/// </summary>
		/// <returns></returns>
		[DebuggerStepThrough]
		private static LogReturnCode RemoveLog()
		{
			LastErrorMessage = string.Empty;

			try
			{
				Debug.Assert(_logFileList != null);
				lock (_logFileList)
				{
					while (_logFileList.Count > (int)_fileCount)
					{
						string removeFilePath = null;
						foreach (string filePath in _logFileList)
						{
							try
							{
								if (removeFilePath == null || File.GetCreationTimeUtc(filePath) < File.GetCreationTimeUtc(removeFilePath))
								{
									removeFilePath = filePath;
								}
							}
							catch (Exception ex)
							{
								LastErrorMessage = ex.Message;
								return LogReturnCode.IOError;
							}
						}
						if (removeFilePath != null)
						{
							try
							{
								File.Delete(removeFilePath);
							}
							catch (Exception ex)
							{
								LastErrorMessage = ex.Message + "\r\nFailed deleting '" + removeFilePath + "'.";
								return LogReturnCode.IOError;
							}
							_logFileList.Remove(removeFilePath);
						}
					}
				}

				return LogReturnCode.Success;
			}
			catch (Exception ex)
			{
				LastErrorMessage = ex.Message;
				return LogReturnCode.UnknownError;
			}
		}

		/// <summary>
		/// Rolls log files.
		/// </summary>
		/// <returns></returns>
		[DebuggerStepThrough]
		private static LogReturnCode RollingFile()
		{
			LastErrorMessage = string.Empty;

			try
			{
				if (string.IsNullOrEmpty(_filePath) || !File.Exists(_filePath))
				{
					return LogReturnCode.FileNotExist;
				}

				long fileSize = new FileInfo(_filePath).Length;

				LogReturnCode ret;
				if ((LogType == LogType.LogDaily && _lastLogDate.Date != DateTime.Now.Date) ||
					(LogType == LogType.LogFileSize && fileSize >= (long)_fileMaxSize))
				{
					string deleteFilePath = _filePath;

					// set new file path
					if ((ret = SetNewFilePath()) != LogReturnCode.Success)
					{
						return ret;
					}

					// remove file
					Debug.Assert(_logFileList != null);
					lock (_logFileList)
					{
						_logFileList.Remove(deleteFilePath);
					}

					try
					{
						File.Delete(deleteFilePath);
					}
					catch (Exception ex)
					{
						LastErrorMessage = ex.Message;
						return LogReturnCode.IOError;
					}
				}
				else if ((LogType == LogType.LogDailyRolling && _lastLogDate.Date != DateTime.Now.Date) ||
					(LogType == LogType.LogFileSizeRolling && fileSize >= (long)_fileMaxSize))
				{
					// set new file path
					if ((ret = SetNewFilePath()) != LogReturnCode.Success)
					{
						return ret;
					}

					// remove logs
					ret = RemoveLog();
					if (ret != LogReturnCode.Success)
					{
						return ret;
					}
				}
			}
			catch (Exception ex)
			{
				LastErrorMessage = ex.Message;
				return LogReturnCode.UnknownError;
			}

			return LogReturnCode.Success;
		}

		/// <summary>
		/// Sets new file path.
		/// </summary>
		/// <param name="appendDateTime">Determines if to append date/time in the file name.</param>
		/// <returns></returns>
		[DebuggerStepThrough]
		private static LogReturnCode SetNewFilePath(bool appendDateTime = true)
		{
			LastErrorMessage = string.Empty;

			lock (_fileWriterObjectLocker)
			{
				if (_fileWriterObject != null)
				{
					try
					{
						_fileWriterObject.Flush();
						_fileWriterObject.Close();
					}
					catch { }
				}

				// Create a StreamWriter
				try
				{
					Debug.Assert(_logFolder != null);
					Debug.Assert(_fileNameWithoutExtension != null);
					Debug.Assert(_fileExtension != null);

					// format file name
					_filePath = Path.Combine(_logFolder, _fileNameWithoutExtension + (appendDateTime ? "_" + DateTime.Now.ToString("yyyy_MM_dd_HH_mm_ss_fff") : "") + _fileExtension);
					_fileWriterObject = new StreamWriter(_filePath, true);
					_fileWriterObject.AutoFlush = true;
				}
				catch (Exception ex)
				{
					LastErrorMessage = ex.Message;
					return LogReturnCode.IOError;
				}

				Debug.Assert(_logFileList != null);
				lock (_logFileList)
				{
					_logFileList.Add(_filePath);
				}
			}

			return LogReturnCode.Success;
		}

		/// <summary>
		/// Gets the full name of the specified type.
		/// </summary>
		/// <param name="type">The type.</param>
		/// <returns></returns>
		[DebuggerStepThrough]
		private static string GetFullName(Type type)
		{
			try
			{
				string fullName = "";
				if (type.IsGenericType)
				{
					Type[] genericParameters = type.GetGenericArguments();
					fullName = GetSpecifiedName(type.Name, ref genericParameters);

					if (type.DeclaringType != null)
					{
						fullName = (type.DeclaringType.ContainsGenericParameters ? GetSpecifiedName(type.DeclaringType.Name, ref genericParameters) : type.DeclaringType.Name) + "." + fullName;
					}
				}
				else
				{
					fullName = type.Name;
					if (type.DeclaringType != null)
					{
						fullName = type.DeclaringType.Name + "." + fullName;
					}
				}

				if (string.IsNullOrEmpty(type.Namespace) == false)
				{
					fullName = type.Namespace + "." + fullName;
				}

				return fullName;
			}
			catch
			{
				return type.FullName;
			}
		}

		[DebuggerStepThrough]
		private static string GetSpecifiedName(string name, ref Type[] additionalGenericParameters)
		{
			int used = 0;
			Match match = Regex.Match(name, "^(.*)`([0-9]+)$", RegexOptions.Compiled);
			if (match.Success)
			{
				used = Convert.ToInt32(match.Groups[2].Value);
				string fullName = match.Groups[1].Value + "<" + string.Join(",", additionalGenericParameters.Skip(additionalGenericParameters.Length - used).Take(used).Select(item => GetFullName(item))) + ">";
				additionalGenericParameters = additionalGenericParameters.Take(additionalGenericParameters.Length - used).ToArray();
				return fullName;
			}
			else
			{
				return name;
			}
		}

		#endregion Static Method
	}

	#endregion EpdaLogger
}
