﻿using System;
using System.Diagnostics;
using System.Linq;

namespace AirFusion.Common
{
	/// <summary>
	/// Represents additional information for a specific element.
	/// </summary>
	[AttributeUsage(AttributeTargets.All, AllowMultiple = false)]
	public sealed class ParameterValue : Attribute
	{
		/// <summary>
		/// Gets or sets the value.
		/// </summary>
		public object Value { get; set; }

		/// <summary>
		/// Initializes a new instance of the <see cref="ParameterValue"/> class.
		/// </summary>
		/// <param name="value">The value.</param>
		public ParameterValue(object value)
		{
			this.Value = value;
		}

		/// <summary>
		/// Returns a <see cref="System.String" /> that represents this instance.
		/// </summary>
		/// <returns>
		/// A <see cref="System.String" /> that represents this instance.
		/// </returns>
		public override string ToString()
		{
			return this.Value != null ? this.Value.ToString() : "(null)";
		}
	}

	/// <summary>
	/// Extensions.
	/// </summary>
	public static partial class Extensions
	{
		/// <summary>
		/// Gets parameter value of the specified enum value.
		/// </summary>
		/// <param name="enumValue">The enum value to get parameter value.</param>
		/// <returns></returns>
		public static object GetParameterValue(this Enum enumValue)
		{
			try
			{
				return enumValue
					.GetType()
					.GetMember(enumValue.ToString()).Where(member => member.MemberType == System.Reflection.MemberTypes.Field)
					.FirstOrDefault()
					.GetCustomAttributes(typeof(ParameterValue), true)
					.OfType<ParameterValue>()
					.Select(item => item.Value)
					.SingleOrDefault();
			}
			catch
			{
				throw new InvalidOperationException("Cannot get 'ParameterValue' attribute.");
			}
		}

		/// <summary>
		/// Tries to get the parameter value.
		/// </summary>
		/// <param name="enumValue">The enum value to get parameter value.</param>
		/// <param name="value">The parameter value.</param>
		/// <returns>True if the field has ParameterValue attribute, false otherwise.</returns>
		public static bool TryGetParameterValue(this Enum enumValue, out object value)
		{
			value = null;
			try
			{
				var parameterValues = enumValue.GetType()
					.GetMember(enumValue.ToString()).Where(member => member.MemberType == System.Reflection.MemberTypes.Field)
					.FirstOrDefault()
					.GetCustomAttributes(typeof(ParameterValue), true)
					.OfType<ParameterValue>().ToArray();
				if (parameterValues != null && parameterValues.Length == 1)
				{
					value = parameterValues.First().Value;
					return true;
				}
				else
				{
					return false;
				}
			}
			catch
			{
				return false;
			}
		}

		/// <summary>
		/// Gets string value of the specified enum value. (first priority: GetParameterValue(), second value: ToString())
		/// </summary>
		/// <param name="enumValue">The enum value to get string value.</param>
		/// <returns></returns>
		public static string GetStringValue(this Enum enumValue)
		{
			try
			{
				object value = enumValue.GetParameterValue();
				if (value is string)
				{
					return value as string;
				}
			}
			catch { }

			return enumValue.ToString();
		}

		/// <summary>
		/// Gets enum value of the specified enum type that has the attribute.
		/// </summary>
		/// <param name="value">The value.</param>
		/// <param name="enumType">The enum type.</param>
		/// <returns></returns>
		public static Enum GetEnumValue(this object value, Type enumType)
		{
			if (enumType.IsEnum == false)
			{
				throw new ArgumentException("The given type is not an enum type.", "enumType");
			}

			foreach (var item in enumType.GetEnumValues())
			{
				Enum enumValue = item as Enum;

				try
				{
					object parameterValue = enumValue.GetParameterValue();

					// try to compare with the original type
					if (value.Equals(parameterValue))
					{
						return enumValue;
					}

					// try to convert the type and compare
					if (parameterValue != null && Convert.ChangeType(value, parameterValue.GetType()).Equals(parameterValue))
					{
						return enumValue;
					}
				}
				catch (Exception ex)
				{
					Debug.WriteLine(ex.Message);
				}
			}

			return null;
		}
	}
}