﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.Common.ProximityDetection.SoundTrigger
{
	/// <summary>
	/// Represents the of callback arguments of SoundTrigger proximity detection.
	/// </summary>
	public class SoundTriggerProximityDetectionCallback : IProximityDetectionCallbackArgs
	{
		private Guid _identity;
		private byte[] _additionalData;

		/// <summary>
		/// Initializes a new instance of the <see cref="SoundTriggerProximityDetectionCallback"/> class.
		/// </summary>
		/// <param name="identity">The identity.</param>
		/// <param name="additionalData">The additional data.</param>
		internal SoundTriggerProximityDetectionCallback(Guid identity, byte[] additionalData)
		{
			this._identity = identity;
			this._additionalData = additionalData;
		}

		/// <summary>
		/// Gets the identity.
		/// </summary>
		/// <returns></returns>
		public virtual Guid GetIdentity()
		{
			return this._identity;
		}

		/// <summary>
		/// Gets the additional data.
		/// </summary>
		/// <returns></returns>
		public virtual byte[] GetAdditionalData()
		{
			/* unmark this if additional data is defined.
			return this._additionalData;
			*/
			throw new NotImplementedException();
		}
	}

	/// <summary>
	/// Represents the proximity detection of SoundTrigger.
	/// </summary>
	public class SoundTriggerProximityDetectionAdapter : IProximityDetection
	{
		#region Definitions or Constants

		private const string SoundTriggerAppKey = "com.acer.airfusion";

		#endregion Definitions or Constants

		#region Private/Protected/Internal Variables

		private ProximityDetectionCallback _callback;
		private bool _isDetecting;

		#endregion Private/Protected/Internal Variables

		#region Interface Methods

		/// <summary>
		/// Starts the detection.
		/// </summary>
		/// <param name="callback">The callback.</param>
		/// <exception cref="System.ArgumentNullException" />
		public virtual void StartDetection(ProximityDetectionCallback callback)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (callback == null)
				{
					throw new ArgumentNullException("callback");
				}

				// start/stop sound listening
				try
				{
					AFSoundTrigger.SoundReceived += this.OnSoundTriggerManagerSoundReceived;
					AFSoundTrigger.RegisterAndStartListen(SoundTriggerAppKey);

					this._callback = callback;
					this._isDetecting = true;
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Failed registering SoundTrigger, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Stops the detection.
		/// </summary>
		public virtual void StopDetection()
		{
			// start/stop sound listening
			try
			{
				this._isDetecting = false;
				this._callback = null;

				AFSoundTrigger.SoundReceived -= this.OnSoundTriggerManagerSoundReceived;
				AFSoundTrigger.UnregisterAndStopListen();
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("Failed unregistering SoundTrigger, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}
		}

		#endregion Interface Methods

		#region Private/Protected/Internal Methods

		internal SoundTriggerProximityDetectionAdapter()
		{
			// do nothing, prevent from instantiation by external
		}

		#endregion Private/Protected/Internal Methods

		#region Callbacks

		/// <summary>
		/// Called when a sound message is received.
		/// </summary>
		/// <param name="soundMessage">The sound message.</param>
		private void OnSoundTriggerManagerSoundReceived(byte[] soundMessage)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("soundMessage: " + (soundMessage != null ? soundMessage.ToHexString() : "N/A"));

				// check parameters
				if (soundMessage == null)
				{
					StaticLogger.LogWarn("The sound message is null, discard.");
				}

				// check prerequisite
				if (!this._isDetecting)
				{
					StaticLogger.LogWarn("It's not discoverable currently, discarded.");
					return;
				}

				// get identity
				Guid identity = Guid.Empty;
				if (soundMessage.Length == 36) // XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX
				{
					string text = null;
					try
					{
						text = Encoding.UTF8.GetString(soundMessage);
						StaticLogger.LogInfo("Received text: " + text);
						if (Guid.TryParse(text, out identity) == false)
						{
							StaticLogger.LogWarn("'" + text + "' is not a valid GUID string.");
							return;
						}
					}
					catch (Exception ex)
					{
						StaticLogger.LogWarn("Failed converting the sound message into UTF-8 string, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
						return;
					}
				}
				else if (soundMessage.Length == 16) // byte array
				{
					try
					{
						identity = new Guid(soundMessage);
					}
					catch (Exception ex)
					{
						StaticLogger.LogWarn("The byte array is not a valid GUID format, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
						return;
					}
				}
				else
				{
					StaticLogger.LogWarn("Unknown message:\r\n" + soundMessage.ToHexString());
					return;
				}

				StaticLogger.LogInfo("identity: " + identity);

				try
				{
					// callback
					if (this._callback != null)
					{
						this._callback(new SoundTriggerProximityDetectionCallback(identity, null)); // there's no additional data defined currently
					}
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Unhandled exception of callback function, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Callbacks
	}
}