﻿using AirFusion.Common;
using AirFusion.Common.Network;
using AirFusion.Common.ProximityDetection;
using AirFusion.WebService;
using AirFusion.WebService.Adapters;
//using InTheHand.Net; // 2015-05-20 Seoker: remove Bluetooth related functions since it's defeatured
using System;
using System.Collections.Generic;
using System.Diagnostics;

namespace AirFusion.Manager.DeviceManager
{
	#region Public Properties

	public sealed partial class UserDeviceManager
	{
		/// <summary>
		/// Gets the device store including discovered device and paired device.
		/// </summary>
		public UserDeviceStore DeviceStore { get; internal set; }

		/// <summary>
		/// Gets the instance of local device with current user.
		/// </summary>
		public AFUserDevice LocalDevice
		{
			// TODO:
			// Make AFUserDevice as abstraction (IDeviceInstance)
			get
			{
				if (this._localDevice == null)
				{
					this.ReadLocalDeviceFixedInformation();
				}

				// update variant properties here
				{
					// web service
					if (this.WebServiceManager != null)
					{
						this._localDevice.WebServicePort = this.WebServiceManager.Port;
					}

					/*
					// update Bluetooth if it's supported
					if (CommonUtils.IsBluetoothEnabled() && (this._localDevice.Bluetooth.Address == null || this._localDevice.Bluetooth.Address.Equals(BluetoothAddress.None)))
					{
						this._localDevice.Bluetooth = new AFUserDevice.BluetoothInfo()
						{
							Address = CommonUtils.GetBluetoothAddress()
						};
					}
					*/
				}

				return this._localDevice;
			}
			internal set
			{
				this._localDevice = value;
			}
		}

		/// <summary>
		/// Gets the profile of local device with current user.
		/// </summary>
		public InstanceProfile LocalProfile
		{
			get
			{
				if (this._localProfile == null && this.LocalDevice != null)
				{
					this._localProfile = this.LocalDevice.ToProfile();
				}

				if (this._localProfile.Network == null)
				{
					try
					{
						var list = new List<InstanceProfile.NetworkInfo>();
						foreach (var networkInfo in NetworkHelper.GetAvailableSupportedNetworkList())
						{
							list.Add(networkInfo.Transform());
						}
						this._localProfile.Network = list.ToArray();
					}
					catch (Exception ex)
					{
						StaticLogger.LogError("Failed getting network information, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
					}
				}

				return this._localProfile;
			}
		}

		/// <summary>
		/// Gets the status that indicates if the discovery is under processing.
		/// </summary>
		public bool IsDiscovering { get; private set; }

		/// <summary>
		/// Gets the status that indicates if the device is able to be discovered.
		/// </summary>
		public bool IsDiscoverable { get; private set; }

		/// <summary>
		/// Gets the configuration.
		/// </summary>
		public DeviceConfiguration Configuration { get; private set; }

		/// <summary>
		/// Gets the web service manager.
		/// </summary>
		internal WebServicesManager WebServiceManager { get; private set; }

		/// <summary>
		/// Gets the device information service.
		/// </summary>
		internal DeviceInformationService DeviceInformationService { get; private set; }

		/// <summary>
		/// Gets the pairing service.
		/// </summary>
		internal PairingService PairingService { get; private set; }

		/// <summary>
		/// Gets the connection helper service.
		/// </summary>
		internal ConnectionHelperService ConnectionHelperService { get; private set; }
	}

	#endregion Public Properties

	#region Events

	public sealed partial class UserDeviceManager
	{
		/// <summary>
		/// Occurs when the initialization is finished.
		/// </summary>
		public event InitializedEventHandler Initialized;

		/// <summary>
		/// Occurs when a device is found/update or lost during discovery stage.
		/// </summary>
		public event DiscoveryEventHandler DiscoveryEvent;

		/// <summary>
		/// Occurs when a paired device is found.
		/// </summary>
		public event DeviceFoundEventHandler DeviceFound;

		/// <summary>
		/// Occurs when a paired device is lost.
		/// </summary>
		public event DeviceLostEventHandler DeviceLost;

		/// <summary>
		/// Occurs when the information of a device is updated.
		/// </summary>
		public event DeviceUpdateEventHandler DeviceUpdate;

		/// <summary>
		/// Occurs when a pairing invitation is received.
		/// </summary>
		public event PairingInvitedEventHandler PairingInvited;

		/// <summary>
		/// Occurs when a pairing request is received.
		/// </summary>
		public event PairingRequestedEventHandler PairingRequested;

		/// <summary>
		/// Occurs when a pairing response is received.
		/// </summary>
		public event PairingResponsedEventHandler PairingResponsed;

		/// <summary>
		/// Occurs when a device is paired.
		/// </summary>
		public event DevicePairedEventHandler DevicePaired;

		/// <summary>
		/// Occurs when a device is un-paired.
		/// </summary>
		public event DeviceUnPairedEventHandler DeviceUnPaired;

#pragma warning disable

		/// <summary>
		/// Occurs when an error occurs.
		/// </summary>
		[Obsolete("This event is not currently in use.")]
		public event ErrorEventHandler ErrorOccur;

#pragma warning restore
	}

	#endregion Events

	#region Public Methods

	/// <summary>
	/// Provides the functionality of device management in Air Fusion.
	/// </summary>
	public sealed partial class UserDeviceManager
	{
		/// <summary>
		/// Initializes a new instance of the <see cref="UserDeviceManager"/> class.
		/// </summary>
		/// <remarks>This instance must be initialized by the main thread.</remarks>
		internal UserDeviceManager()
		{
			this.DeviceStore = new UserDeviceStore(this);
			this.IsDiscovering = false;
			this.IsDiscoverable = false;
			this.Configuration = null;
			this.WebServiceManager = null;
			this.DeviceInformationService = null;
			this.PairingService = null;
			this.ConnectionHelperService = null;
			this._initialized = false;
			this._mdnsAdapter = new MulticastDnsAdapter(ServiceName);  // this instance must be initialized by the main thread.
			this._dandelionAdapter = null;
			this._localDevice = null;
			this._localProfile = null;
			this._networkObserver = null;
			//this._bluetoothListener = null; // 2015-05-20 Seoker: remove Bluetooth related functions since it's defeatured
			this._discoveryTimer = null;
			this._proximityDetection = null;
		}

		/// <summary>
		/// Initializes the <see cref="UserDeviceManager"/> object.
		/// </summary>
		/// <param name="parameter">The initial parameters.</param>
		internal void Initialize(InitialParameters parameter)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("parameter: " + (parameter != null ? "\r\n" + parameter.ToString() : "N/A"));

				// check parameters
				if (parameter == null)
				{
					throw new ArgumentNullException("parameter");
				}

				// check prerequisite
				if (this._initialized)
				{
					throw new InvalidOperationException("This instance has been initialized.");
				}

				var stopWatch = new Stopwatch();
				stopWatch.Start();

				try
				{
					// initialize local device
					this.InitializeLocalDevice();

					// initialize web service (this should be done before mDNS/Dandelion adapters being created to ensure the port number is valid
					this.InitializeWebServices(parameter.WebServiceManager);

					// initialize MulticastDnsAdapter
					// TODO: create an common interface for both MulticastDnsAdapter & DandelionAdapter => real adapter pattern
					this.InitializeMulticastDnsAdapter();

					// initialize DandelionAdapter
					// TODO: create an common interface for both MulticastDnsAdapter & DandelionAdapter => real adapter pattern
					this.InitializeDandelionAdapter();

					// initialize Bluetooth
					// 2015-05-20 Seoker: remove Bluetooth related functions since it's been defeatured
					//this.InitializeBluetooth();

					// initialize network observer
					this.InitializeNetworkObserver();

					this._initialized = true;

					stopWatch.Stop();

					// fire Initialized event
					if (this.Initialized != null)
					{
						StaticLogger.LogTrace("Fire Initialized event...");
						CommonUtils.InvokeEvent<InitializedEventArgs>(new EventHandler<InitializedEventArgs>(this.Initialized), this, new InitializedEventArgs(stopWatch.ElapsedMilliseconds));
					}
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Failed initializing, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
					throw;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Enables the discoverability.
		/// </summary>
		/// <param name="proximityDetectionMethod">The proximity detection method.</param>
		/// <exception cref="System.InvalidOperationException" />
		public void EnableDiscoverable(ProximityDetectionMethod proximityDetectionMethod = ProximityDetectionMethod.None)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("proximityDetectionMethod: " + proximityDetectionMethod);

				// check prequisite
				if (this._initialized == false)
				{
					throw new InvalidOperationException("This instance is not initialized yet.");
				}

				lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNDiscoverable)))
				{
					if (this.IsDiscoverable)
					{
						return;
					}
					// start the proximity detection if specified
					if (proximityDetectionMethod != ProximityDetectionMethod.None)
					{
						this._proximityDetection = ProximityDetectionFactory.GetInstance(proximityDetectionMethod);
						this._proximityDetection.StartDetection(this.OnProximityDetectionCallback);
					}

					// start Dandelion advertisement
					this.StartDandelionAdvertise();

					// start mDNS advertisement
					this.StartMulticastDnsAdvertise();

					this.IsDiscoverable = true;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Disables the discoverability.
		/// </summary>
		/// <exception cref="System.InvalidOperationException" />
		public void DisableDiscoverable()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prequisite
				if (this._initialized == false)
				{
					throw new InvalidOperationException("This instance is not initialized yet.");
				}

				lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNDiscoverable)))
				{
					if (!this.IsDiscoverable)
					{
						return;
					}

					this.IsDiscoverable = false;

					// stop the proximity detection if any
					if (this._proximityDetection != null)
					{
						this._proximityDetection.StopDetection();
						this._proximityDetection = null;
					}

					// stop Dandelion advertisement
					this.StopDandelionAdvertise();

					// stop mDNS advertisement
					this.StopMulticastDnsAdvertise();
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Starts discovery.
		/// </summary>
		/// <param name="duration">The duration in milliseconds.</param>
		/// <param name="proximityDetectionMethod">The proximity detection method. It's ignored currently.</param>
		/// <exception cref="System.InvalidOperationException" />
		public void StartDiscovery(double duration = 0.0, ProximityDetectionMethod proximityDetectionMethod = ProximityDetectionMethod.None)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("timeout: " + duration);
				StaticLogger.LogDebug("proximityDetectionMethod: " + proximityDetectionMethod);

				// check prequisite
				if (this._initialized == false)
				{
					throw new InvalidOperationException("This instance is not initialized yet.");
				}

				lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNDiscovery)))
				{
					if (this.IsDiscovering)
					{
						return;
					}

					this.IsDiscovering = true;

					// mDNS
					this.StartMulticastDnsDiscovery(duration);

					// Dandelion
					this.StartDandelionDiscovery();

					// start timer
					{
						// make sure there's no another timer running
						if (this._discoveryTimer != null)
						{
							this._discoveryTimer.Dispose();
							this._discoveryTimer = null;
						}

						// set timer to stop if duration > 0
						if (duration > 0)
						{
							this._discoveryTimer = new System.Timers.Timer(duration);
							this._discoveryTimer.AutoReset = false;
							this._discoveryTimer.Elapsed += (s, e) =>
								{
									try
									{
										this.StopDiscovery();

										if (this._discoveryTimer != null)
										{
											this._discoveryTimer.Dispose();
											this._discoveryTimer = null;
										}
									}
									catch (Exception ex)
									{
										StaticLogger.LogError("Failed to stop discovery, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
									}
								};
							this._discoveryTimer.Start();
						}
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Stops the discovery.
		/// </summary>
		/// <exception cref="System.InvalidOperationException" />
		public void StopDiscovery()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prequisite
				if (this._initialized == false)
				{
					throw new InvalidOperationException("This instance is not initialized yet.");
				}

				lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNDiscovery)))
				{
					if (!this.IsDiscovering)
					{
						return;
					}

					this.IsDiscovering = false;

					// kill timer
					if (this._discoveryTimer != null)
					{
						this._discoveryTimer.Dispose();
						this._discoveryTimer = null;
					}

					// mDNS
					this.StopMulticastDnsDiscovery();

					// Dandelion
					this.StopDandelionDiscovery();
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}
	}

	#endregion Public Methods

	#region Disposable

	public sealed partial class UserDeviceManager : IDisposable
	{
		internal bool _disposed = false;

		internal void Dispose(bool disposing)
		{
			if (!this._disposed)
			{
				// If disposing equals true, dispose all managed and unmanaged resources.
				if (disposing)
				{
					// 2015-05-20 Seoker: remove Bluetooth related functions since it's been defeatured
					/*
					try
					{
						this.StopBluetoothListener();
					}
					catch { }
					*/

					try
					{
						this.StopDiscovery();
					}
					catch { }

					try
					{
						this.DisableDiscoverable();
					}
					catch { }

					lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNMulticastDnsAdapter)))
					{
						if (this._mdnsAdapter != null)
						{
							this._mdnsAdapter.Dispose();
							this._mdnsAdapter = null;
						}
					}

					lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNNetworkObserver)))
					{
						if (this._networkObservable != null)
						{
							this._networkObservable.Dispose();
							this._networkObservable = null;
						}
					}
				}

				//
				// cleanup unmanaged resources
				//

				this._disposed = true;
			}
		}

		/// <summary>
		/// Disposes the object.
		/// </summary>
		public void Dispose()
		{
			this.Dispose(true);

			// This object will be cleaned up by the Dispose method.
			// Therefore, you should call GC.SupressFinalize to take this object off the finalization queue
			// and prevent finalization code for this object from executing a second time.
			GC.SuppressFinalize(this);
		}
	}

	#endregion Disposable
}
