﻿using System;
using System.Collections;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.Manager.DeviceManager
{
	/// <summary>
	/// Represents the store of devices.
	/// </summary>
	public sealed class UserDeviceStore : IReadOnlyCollection<AFUserDevice>
	{
		#region Private/Protected/Internal Variables

		private ConcurrentDictionary<Guid, AFUserDevice> _devicePool;
		private UserDeviceManager _deviceManager;

		#endregion Private/Protected/Internal Variables

		#region Properties

		/// <summary>
		/// Gets particular device by the specified instance ID.
		/// </summary>
		/// <param name="instanceID">The instance ID.</param>
		/// <returns></returns>
		public AFUserDevice this[Guid instanceID]
		{
			get
			{
				Debug.Assert(this._devicePool != null);
				return this._devicePool.ContainsKey(instanceID) ? this._devicePool[instanceID] : null;
			}
			internal set
			{
				Debug.Assert(this._devicePool != null);
				this._devicePool[instanceID] = value;
			}
		}

		#endregion Properties

		#region IReadOnlyCollection<T>, IEnumerable<T>, IEnumerable

		/// <summary>
		/// Gets the number of device.
		/// </summary>
		public int Count
		{
			get
			{
				return this._devicePool != null ? this._devicePool.Count : 0;
			}
		}

		/// <summary>
		/// Returns an enumerator that iterates through the collection.
		/// </summary>
		IEnumerator<AFUserDevice> IEnumerable<AFUserDevice>.GetEnumerator()
		{
			return this._devicePool.Values.GetEnumerator();
		}

		/// <summary>
		/// Returns an enumerator that iterates through a collection.
		/// </summary>
		IEnumerator IEnumerable.GetEnumerator()
		{
			return this._devicePool.Values.GetEnumerator();
		}

		#endregion IReadOnlyCollection<T>, IEnumerable<T>, IEnumerable

		#region Public Methods

		/// <summary>
		/// Gets particular device by the specified instance ID.
		/// </summary>
		/// <param name="instanceID">The instance ID.</param>
		/// <returns></returns>
		public AFUserDevice GetDevice(Guid instanceID)
		{
			return this[instanceID];
		}

		/// <summary>
		/// Determines if the particular device exists with the specified instance ID.
		/// </summary>
		/// <param name="instanceID">The instance ID.</param>
		/// <returns></returns>
		public bool ExistsDevice(Guid instanceID)
		{
			Debug.Assert(this._devicePool != null);
			return this._devicePool.ContainsKey(instanceID);
		}

		/// <summary>
		/// Gets the device by address.
		/// </summary>
		/// <param name="address">The address.</param>
		/// <returns></returns>
		public AFUserDevice GetDeviceByAddress(IPAddress address)
		{
			try
			{
				Debug.Assert(this._devicePool != null);
				return this._devicePool.Where(entry => entry.Value != null && entry.Value.Address.Equals(address)).FirstOrDefault().Value;
			}
			catch
			{
				return null;
			}
		}

		/// <summary>
		/// Gets the device collection by the given condition.
		/// </summary>
		/// <param name="predict">A function to test each element for a condition.</param>
		/// <returns></returns>
		public AFUserDevice[] GetDevicesByCondition(Func<AFUserDevice, bool> predict)
		{
			Debug.Assert(this._devicePool != null);
			lock (this._devicePool)
			{
				try
				{
					return this._devicePool.Where(entry => predict(entry.Value)).Select(entry => entry.Value).ToArray();
				}
				catch (Exception ex)
				{
					throw new Exception("There's unhandled exception from the predict function.", ex);
				}
			}
		}

		/// <summary>
		/// Gets all the devices.
		/// </summary>
		/// <returns></returns>
		public AFUserDevice[] GetAllDevices()
		{
			Debug.Assert(this._devicePool != null);
			return this._devicePool.Values.ToArray();
		}

		/// <summary>
		/// Gets paired devices.
		/// </summary>
		/// <returns></returns>
		public AFUserDevice[] GetPairedDevices()
		{
			return this.GetDevicesByCondition(device => device.PairStatus == AFUserDevice.PairingStatus.Paired).ToArray();
		}

		/// <summary>
		/// Gets non-paired devices.
		/// </summary>
		/// <returns></returns>
		public AFUserDevice[] GetNonPairedDevices()
		{
			return this.GetDevicesByCondition(device => device.PairStatus != AFUserDevice.PairingStatus.Paired).ToArray();
		}

		#endregion Public Methods

		#region Internal Functions

		/// <summary>
		/// Initializes a new instance of the <see cref="UserDeviceStore"/> class.
		/// </summary>
		internal UserDeviceStore(UserDeviceManager deviceManager)
		{
			this._devicePool = new ConcurrentDictionary<Guid, AFUserDevice>();
			this._deviceManager = deviceManager;
		}

		/// <summary>
		/// Gets the device by instance ID or creates one if such device doesn't exist.
		/// </summary>
		/// <param name="instanceID">The instance ID.</param>
		/// <returns></returns>
		internal AFUserDevice GetOrCreateDevice(Guid instanceID)
		{
			Debug.Assert(this._devicePool != null);

			AFUserDevice device = this._devicePool.GetOrAdd(instanceID, new AFUserDevice() { InstanceID = instanceID });

			device.Owner = this._deviceManager;
			device.RegisterPairEventHandler(this._deviceManager.OnDevicePaired);
			device.RegisterUnPairEventHandler(this._deviceManager.OnDeviceUnPaired);

			return device;
		}

		/// <summary>
		/// Removes the specified device from the device list.
		/// </summary>
		/// <param name="instanceID">The instance ID.</param>
		internal void RemoveByInstanceID(Guid instanceID)
		{
			AFUserDevice device;
			Debug.Assert(this._devicePool != null);
			this._devicePool.TryRemove(instanceID, out device);
		}

		/// <summary>
		/// Removes devices by the given condition.
		/// </summary>
		/// <param name="predict">A function to test each element for a condition.</param>
		internal void RemoveByCondition(Func<AFUserDevice, bool> predict)
		{
			// debug assertion
			Debug.Assert(this._devicePool != null);
			try
			{
				var keyList = this._devicePool.Where(entry => predict(entry.Value)).Select(pair => pair.Key).ToArray();

				foreach (var key in keyList)
				{
					AFUserDevice device;
					this._devicePool.TryRemove(key, out device);
				}
			}
			catch { }
		}

		/// <summary>
		/// Adds the specified device into device list or updates the existing one.
		/// </summary>
		/// <param name="device">The device.</param>
		internal AFUserDevice AddOrUpdate(AFUserDevice device)
		{
			// debug assertion
			Debug.Assert(this._devicePool != null);

			// check parameters
			if (device == null)
			{
				throw new ArgumentNullException("device");
			}

			AFUserDevice retDevice = this._devicePool.AddOrUpdate(device.InstanceID, device, delegate(Guid instanceId, AFUserDevice oldDevice)
			{
				oldDevice.FillFromIfSpecified(device);
				return oldDevice;
			});

			device.Owner = this._deviceManager;
			device.RegisterPairEventHandler(this._deviceManager.OnDevicePaired);
			device.RegisterUnPairEventHandler(this._deviceManager.OnDeviceUnPaired);

			return retDevice;
		}

		#endregion Internal Functions
	}
}
