﻿using AirFusion.Common;
using AirFusion.Common.RegistryHolders;
using AirFusion.WebService;
using InTheHand.Net;
using System;
using System.Diagnostics;
using System.Text;

namespace AirFusion.Manager.DeviceManager
{
	public sealed partial class AFUserDevice
	{
		/// <summary>
		/// Represents the Bluetooth information of the device.
		/// </summary>
		public sealed partial class BluetoothInfo
		{
			/// <summary>
			/// Gets the Bluetooth address.
			/// </summary>
			[RegistryStoreTypeAttribute(StoreType.Value), RegistryStoreConverterType(typeof(AirFusion.Common.Converters.BluetoothAddressConverter))]
			public BluetoothAddress Address { get; internal set; }

			/// <summary>
			/// Initializes a new instance of the <see cref="BluetoothInfo"/> class.
			/// </summary>
			internal BluetoothInfo()
			{
				this.Address = null;
			}

			/// <summary>
			/// Fills from the specified <see cref="BluetoothInfo"/> object if such values are specified.
			/// </summary>
			/// <param name="source">The source object.</param>
			internal void FillFromIfSpecified(BluetoothInfo source)
			{
				Debug.Assert(source != null);

				if (source.Address != null && !source.Address.Equals(BluetoothAddress.None))
				{
					this.Address = source.Address;
				}
			}

			/// <summary>
			/// Fills from the specified <see cref="DeviceInformation.BluetoothInfo"/> object if such values are specified.
			/// </summary>
			/// <param name="source">The source object.</param>
			internal void FillFromIfSpecified(DeviceInformation.BluetoothInfo source)
			{
				Debug.Assert(source != null);

				if (source.Address != null && !source.Address.Equals(BluetoothAddress.None))
				{
					this.Address = source.Address;
				}
			}

			/// <summary>
			/// Fills the values into the specified <see cref="DeviceInformation.BluetoothInfo"/> object.
			/// </summary>
			/// <param name="target">The target object.</param>
			internal void FillTo(DeviceInformation.BluetoothInfo target)
			{
				Debug.Assert(target != null);

				target.Address = this.Address;
			}

			#region Override

			/// <summary>
			/// Determines whether the specified object is equal to the current object.
			/// </summary>
			/// <param name="obj">The object to compare with the current object.</param>
			/// <returns></returns>
			public override bool Equals(object obj)
			{
				return (obj is BluetoothInfo && this == (BluetoothInfo)obj);
			}

			/// <summary>
			/// Returns the hash code of the instance.
			/// </summary>
			/// <returns></returns>
			public override int GetHashCode()
			{
				return this.Address != null ? this.Address.GetHashCode() : base.GetHashCode();
			}

			/// <summary>
			/// Implements the operator ==.
			/// </summary>
			/// <param name="x">The first object.</param>
			/// <param name="y">The second object.</param>
			public static bool operator ==(BluetoothInfo x, BluetoothInfo y)
			{
				return (((object)x == null && (object)y == null) || Object.ReferenceEquals(x, y) ||
					((object)x != null && (object)y != null) &&
					x.Address == y.Address);
			}

			/// <summary>
			/// Implements the operator !=.
			/// </summary>
			/// <param name="x">The first object.</param>
			/// <param name="y">The second object.</param>
			public static bool operator !=(BluetoothInfo x, BluetoothInfo y)
			{
				return !(x == y);
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + ".Address: " + (this.Address != null ? this.Address.GetStringFormat() : "N/A"));

				return builder.ToString().TrimEnd();
			}

			#endregion Override
		}
	}
}