﻿using AirFusion.Common;
using AirFusion.Common.RegistryHolders;
using AirFusion.WebService.Adapters;
using System;
using System.Diagnostics;
using System.Text;

namespace AirFusion.Manager.DeviceManager
{
	public sealed partial class AFUserDevice
	{
		public sealed partial class ServicesInfo
		{
			/// <summary>
			/// Represents the information of the input control.
			/// </summary>
			public sealed class InputControlInfo
			{
				/// <summary>
				/// Gets or sets the port number of the VNC server component.
				/// </summary>
				[RegistryStoreTypeAttribute(StoreType.Value)]
				public ushort? Port { get; private set; }

				/// <summary>
				/// Sets the port number.
				/// </summary>
				public void SetPort(ushort port)
				{
					this.Port = port;
				}

				/// <summary>
				/// Fills from the specified <see cref="InputControlInfo"/> object if such values are specified.
				/// </summary>
				/// <param name="source">The source object.</param>
				internal void FillFromIfSpecified(InputControlInfo source)
				{
					Debug.Assert(source != null);

					if (source.Port.HasValue)
					{
						this.Port = source.Port;
					}
				}

				/// <summary>
				/// Fills from the specified <see cref="DeviceInformation.ServicesInfo.InputControlInfo"/> object if such values are specified.
				/// </summary>
				/// <param name="source">The source object.</param>
				internal void FillFromIfSpecified(DeviceInformation.ServicesInfo.InputControlInfo source)
				{
					Debug.Assert(source != null);

					if (source.Port.HasValue)
					{
						this.Port = source.Port;
					}
				}

				/// <summary>
				/// Fills the values into the specified <seealso cref="DeviceInformation.ServicesInfo.InputControlInfo"/> object.
				/// </summary>
				/// <param name="target">The target object.</param>
				internal void FillTo(DeviceInformation.ServicesInfo.InputControlInfo target)
				{
					Debug.Assert(target != null);

					target.Port = this.Port;
				}

				#region Override

				/// <summary>
				/// Determines whether the specified object is equal to the current object.
				/// </summary>
				/// <param name="obj">The object to compare with the current object.</param>
				/// <returns></returns>
				public override bool Equals(object obj)
				{
					return (obj is InputControlInfo && this == (InputControlInfo)obj);
				}

				/// <summary>
				/// Returns the hash code of the instance.
				/// </summary>
				/// <returns></returns>
				public override int GetHashCode()
				{
					return base.GetHashCode() ^ (this.Port.HasValue ? this.Port.GetHashCode() : 0);
				}

				/// <summary>
				/// Implements the operator ==.
				/// </summary>
				/// <param name="x">The first object.</param>
				/// <param name="y">The second object.</param>
				public static bool operator ==(InputControlInfo x, InputControlInfo y)
				{
					return (((object)x == null && (object)y == null) || Object.ReferenceEquals(x, y) ||
						((object)x != null && (object)y != null) &&
						x.Port == y.Port);
				}

				/// <summary>
				/// Implements the operator !=.
				/// </summary>
				/// <param name="x">The first object.</param>
				/// <param name="y">The second object.</param>
				public static bool operator !=(InputControlInfo x, InputControlInfo y)
				{
					return !(x == y);
				}

				/// <summary>
				/// Returns a <see cref="System.String" /> that represents this instance.
				/// </summary>
				/// <returns>
				/// A <see cref="System.String" /> that represents this instance.
				/// </returns>
				public override string ToString()
				{
					StringBuilder builder = new StringBuilder();
					builder.AppendLine(this.GetType().GetFullName() + ".Port: " + (this.Port.HasValue ? this.Port.Value.ToString() : "N/A"));

					return builder.ToString().TrimEnd();
				}

				#endregion Override
			}
		}
	}
}
