﻿using AirFusion.Common;
using AirFusion.Common.RegistryHolders;
using AirFusion.WebService.Adapters;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;

namespace AirFusion.Manager.DeviceManager
{
	public sealed partial class AFUserDevice
	{
		public sealed partial class ServicesInfo
		{
			/// <summary>
			/// Represents the information of the screencast component.
			/// </summary>
			public sealed partial class ScreencastInfo
			{
				/// <summary>
				/// Gets or sets the port number of the RTSP port.
				/// </summary>
				[RegistryStoreTypeAttribute(StoreType.Value)]
				public ushort? Port { get; private set; }

				/// <summary>
				/// Gets the supported resolutions.
				/// </summary>
				/// <remarks>The property might be null.</remarks>
				[RegistryStoreTypeAttribute(StoreType.Value)]
				public IList<ScreenResolution> SupportedResolutions { get; internal set; }

				/// <summary>
				/// Initializes a new instance of the <see cref="ScreencastInfo"/> class.
				/// </summary>
				internal ScreencastInfo()
				{
					this.Port = null;
					this.SupportedResolutions = null;
				}

				/// <summary>
				/// Adds the specified resolution into supported resolutions.
				/// </summary>
				/// <param name="resolution">The resolution.</param>
				public void AddSupportedResolution(ScreenResolution resolution)
				{
					lock (CommonUtils.GetLocker(new Tuple<ScreencastInfo, string>(this, "SupportedResolutions")))
					{
						if (this.SupportedResolutions == null)
						{
							this.SupportedResolutions = new List<ScreenResolution>();
						}

						if (this.SupportedResolutions.Contains(resolution) == false)
						{
							this.SupportedResolutions.Add(resolution);
						}
					}
				}

				/// <summary>
				/// Removes specified resolution from supported resolutions.
				/// </summary>
				/// <param name="resolution">The resolution.</param>
				public void RemoveSupportedResolution(ScreenResolution resolution)
				{
					lock (CommonUtils.GetLocker(new Tuple<ScreencastInfo, string>(this, "SupportedResolutions")))
					{
						if (this.SupportedResolutions != null)
						{
							this.SupportedResolutions.Remove(resolution);
						}
					}
				}

				/// <summary>
				/// Sets the port number.
				/// </summary>
				public void SetPort(ushort port)
				{
					this.Port = port;
				}

				/// <summary>
				/// Fills from the specified <see cref="ScreencastInfo"/> object if such values are specified.
				/// </summary>
				/// <param name="source">The source object.</param>
				internal void FillFromIfSpecified(ScreencastInfo source)
				{
					Debug.Assert(source != null);

					if (source.Port.HasValue)
					{
						this.Port = source.Port;
					}

					if (source.SupportedResolutions != null)
					{
						this.SupportedResolutions = source.SupportedResolutions.ToList(); // a tricky: to clone the list
					}
				}

				/// <summary>
				/// Fills from the specified <see cref="DeviceInformation.ServicesInfo.ScreencastInfo"/> object if such values are specified.
				/// </summary>
				/// <param name="source">The source object.</param>
				internal void FillFromIfSpecified(DeviceInformation.ServicesInfo.ScreencastInfo source)
				{
					Debug.Assert(source != null);

					if (source.Port.HasValue)
					{
						this.Port = source.Port;
					}

					if (source.SupportedResolutions != null)
					{
						this.SupportedResolutions = source.SupportedResolutions
							.Where(item => item.Width.HasValue && item.Height.HasValue)
							.Select(item => new ScreenResolution(item.Width.Value, item.Height.Value)).ToList();
					}
				}

				/// <summary>
				/// Fills the values into the specified <see cref="DeviceInformation.ServicesInfo.ScreencastInfo"/> object.
				/// </summary>
				/// <param name="target">The target object.</param>
				internal void FillTo(DeviceInformation.ServicesInfo.ScreencastInfo target)
				{
					Debug.Assert(target != null);

					target.Port = this.Port;

					if (this.SupportedResolutions != null)
					{
						target.SupportedResolutions = this.SupportedResolutions
							.Select(item => new DeviceInformation.ServicesInfo.ScreencastInfo.ScreenResolution(item.Width, item.Height)).ToArray();
					}
					else
					{
						target.SupportedResolutions = null;
					}
				}

				#region Override

				/// <summary>
				/// Determines whether the specified object is equal to the current object.
				/// </summary>
				/// <param name="obj">The object to compare with the current object.</param>
				/// <returns></returns>
				public override bool Equals(object obj)
				{
					return (obj is ScreencastInfo && this == (ScreencastInfo)obj);
				}

				/// <summary>
				/// Returns the hash code of the instance.
				/// </summary>
				/// <returns></returns>
				public override int GetHashCode()
				{
					return base.GetHashCode() ^ (this.Port.HasValue ? this.Port.GetHashCode() : 0);
				}

				/// <summary>
				/// Implements the operator ==.
				/// </summary>
				/// <param name="x">The first object.</param>
				/// <param name="y">The second object.</param>
				public static bool operator ==(ScreencastInfo x, ScreencastInfo y)
				{
					return (((object)x == null && (object)y == null) || Object.ReferenceEquals(x, y) ||
						((object)x != null && (object)y != null) &&
						x.Port == y.Port &&
						((x.SupportedResolutions == null) == (y.SupportedResolutions == null)) &&
						((x.SupportedResolutions == null && y.SupportedResolutions == null) ||
						x.SupportedResolutions.SequenceEqual(y.SupportedResolutions)));
				}

				/// <summary>
				/// Implements the operator !=.
				/// </summary>
				/// <param name="x">The first object.</param>
				/// <param name="y">The second object.</param>
				public static bool operator !=(ScreencastInfo x, ScreencastInfo y)
				{
					return !(x == y);
				}

				/// <summary>
				/// Returns a <see cref="System.String" /> that represents this instance.
				/// </summary>
				/// <returns>
				/// A <see cref="System.String" /> that represents this instance.
				/// </returns>
				public override string ToString()
				{
					StringBuilder builder = new StringBuilder();
					builder.AppendLine(this.GetType().GetFullName() + ".Port: " + (this.Port.HasValue ? this.Port.Value.ToString() : "N/A"));
					builder.AppendLine(this.GetType().GetFullName() + ".SupportedResolutions: " + (this.SupportedResolutions != null ? string.Join("; ", this.SupportedResolutions) : "N/A"));

					return builder.ToString().TrimEnd();
				}

				#endregion Override
			}
		}
	}
}
