﻿using AirFusion.Common;
using AirFusion.Common.RegistryHolders;
using AirFusion.WebService.Adapters;
using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Threading.Tasks;

namespace AirFusion.Manager.DeviceManager
{
	// TODO: rename AFUserDevice to UserDevice or DeviceInstance or Device
	/// <summary>
	/// Represents the user device.
	/// </summary>
	public sealed partial class AFUserDevice : INotifyPropertyChanged, IEquatable<AFUserDevice>
	{
		#region Properties

		#region Generic

		/// <summary>
		/// Gets the instance ID.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.Value)]
		public Guid InstanceID { get; internal set; }

		/// <summary>
		/// Gets the pairing session ID.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.SubTree)] // need converter
		public ISessionId SessionID { get; internal set; }

		#endregion Generic

		#region Device-specific

		/// <summary>
		/// Gets the device name.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.Value)]
		public string DeviceName { get; internal set; }

		/// <summary>
		/// Gets the model name.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.Value)]
		public string ModelName { get; internal set; }

		/// <summary>
		/// Gets the manufacturer of the device.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.Value)]
		public string Manufacturer { get; internal set; }

		/// <summary>
		/// Gets the OS type of the device.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.Value)]
		public OSTypeValue OSType { get; internal set; }

		/// <summary>
		/// Gets the operating system of the device.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.Value)]
		public string OS { get; internal set; }

		/// <summary>
		/// Gets the version of the operating system of the device.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.Value)]
		public string OSVersion { get; internal set; }

		/// <summary>
		/// Gets the serial number.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.Value)]
		public string SerialNumber { get; internal set; }

		/// <summary>
		/// Gets the display information.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.SubTree)]
		public DisplayInfo Display { get; internal set; }

		// 2015-05-20 Seoker: remove Bluetooth related functions since it's been defeatured
		/*
		/// <summary>
		/// Gets the information of Bluetooth.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.SubTree)]
		public BluetoothInfo Bluetooth { get; internal set; }
		*/

		#endregion Device-specific

		#region User-specific

		/// <summary>
		/// Gets the user name.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.Value)]
		public string UserName { get; internal set; }

		#endregion User-specific

		#region App-specific

		/// <summary>
		/// Gets the application version.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.Value)]
		public string ApplicationVersion { get; internal set; }

		/// <summary>
		/// Gets or sets the value indicating whether the application has system permission.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.Value)]
		public bool? HasSystemPermission { get; internal set; }

		/// <summary>
		/// Gets the information of the AirFusion services.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.SubTree)]
		public ServicesInfo Services { get; internal set; }

		#endregion App-specific

		#region Miscellaneous

		/// <summary>
		/// Gets the indication that indicates if the user device is local device.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.NoStore)]
		public bool IsLocalDevice
		{
			get
			{
				return this.Owner != null && Object.ReferenceEquals(this.Owner.LocalDevice, this);
			}
		}

		/// <summary>
		/// Gets or sets the pair status of the device. Specify true to keep the device afterwards.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.NoStore)]
		public PairingStatus PairStatus
		{
			get
			{
				return this._pairStatus;
			}
			internal set
			{
				if (this._pairStatus != value)
				{
					bool fireEvent = !(this._pairStatus == PairingStatus.PairingPending && value == PairingStatus.UnPaired);
					this._pairStatus = value;

					if (fireEvent)
					{
						if (value == PairingStatus.Paired)
						{
							// fire Paired event
							if (this.Paired != null)
							{
								StaticLogger.LogTrace("Fire Paired event...");
								CommonUtils.InvokeEvent<DevicePairedEventArgs>(
									new EventHandler<DevicePairedEventArgs>(this.Paired),
									this,
									new DevicePairedEventArgs(this));
							}
						}
						else if (value == PairingStatus.UnPaired)
						{
							// fire UnPaired event
							if (this.UnPaired != null)
							{
								StaticLogger.LogTrace("Fire UnPaired event...");
								CommonUtils.InvokeEvent<DeviceUnPairedEventArgs>(
									new EventHandler<DeviceUnPairedEventArgs>(this.UnPaired),
									this,
									new DeviceUnPairedEventArgs(this));
							}
						}
					}
				}
			}
		}

		/// <summary>
		/// Gets the address of the device.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.Value)]
		public IPAddress Address { get; internal set; }

		/// <summary>
		/// Gets the port number of the web service of the user device.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.Value), RegistryStoreName("Port")]
		public ushort WebServicePort
		{
			get
			{
				if (this._webServicePort.HasValue && this._webServicePort != 0)
				{
					return this._webServicePort.Value;
				}

				StaticLogger.LogDebug("The service port number is " + (this._webServicePort.HasValue ? "unset" : "zero") + ", use default one '" + AirFusionManager.WebServicePort[0] + "' for workaround.");
				Debugger.Break();
				return AirFusionManager.WebServicePort[0];
			}
			internal set
			{
				this._webServicePort = value;
			}
		}

		/// <summary>
		/// Gets the friendly name of the device.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.NoStore)]
		public string FriendlyName
		{
			get
			{
				string friendlyName = this.DeviceName;
				if (string.IsNullOrEmpty(friendlyName))
				{
					friendlyName = this.ModelName;
				}
				if (string.IsNullOrEmpty(friendlyName) && this.Address != null && this.Address.Equals(IPAddress.None) == false)
				{
					friendlyName = this.Address.ToString();
				}
				if (string.IsNullOrEmpty(friendlyName) && this.InstanceID.CompareTo(Guid.Empty) == 0)
				{
					friendlyName = this.InstanceID.GetStringFormat();
				}
				if (string.IsNullOrEmpty(friendlyName))
				{
					friendlyName = "Unknown";
				}

				return friendlyName;
			}
		}

		/// <summary>
		/// Gets the owned <see cref="UserDeviceManager"/>.
		/// </summary>
		[RegistryStoreTypeAttribute(StoreType.NoStore)]
		public UserDeviceManager Owner { get; internal set; }

		#endregion Miscellaneous

		#endregion Properties

		#region Events

		/// <summary>
		/// Occurs when the device is paired.
		/// </summary>
		internal event UserDeviceManager.DevicePairedEventHandler Paired;

		/// <summary>
		/// Occurs when the device is unpaired.
		/// </summary>
		internal event UserDeviceManager.DeviceUnPairedEventHandler UnPaired;

		#endregion Events

		#region Public Methods

		/// <summary>
		/// Unpairs the device.
		/// </summary>
		public void UnPair()
		{
			try
			{
				StaticLogger.TrackEnter();

				this.PairStatus = PairingStatus.UnPaired;

				// update the pairing status to remote device
				this.UpdatePairingStatusAsync().ContinueWith((t) =>
				{
					if (t.IsFaulted)
					{
						StaticLogger.LogWarn("Failed updating pairing status, exception:\r\n" + CommonUtils.GetExceptionMessages(t.Exception));
					}
				}, TaskContinuationOptions.ExecuteSynchronously);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Wakes up the device asynchronously.
		/// </summary>
		/// <returns>A value indicating whether the operation is done successfully.</returns>
		public async Task<bool> WakeUpAsync()
		{
			try
			{
				StaticLogger.TrackEnter();

				var taskList = this.GetOrCreateWakeUpTasks().Values.ToList();

				// wait until any approach success
				while (taskList != null && taskList.Count > 0)
				{
					try
					{
						var task = await Task<bool>.WhenAny(taskList);
						if (task.IsCompleted)
						{
							taskList.Remove(task);
							if (task.Result == true)
							{
								return true;
							}
						}
					}
					catch (AggregateException ae)
					{
						foreach (var e in ae.Flatten().InnerExceptions)
						{
							if (e is WebServiceVersionNotMatchedException)
							{
								// Recover from the exception. Here we just
								// print the message for demonstration purposes.
								StaticLogger.LogError("WakeUpAsynce Version'" + this + "': " + e.Message);
								throw e;
							}
						}
					}
				}

				return false;
			}
			catch (WebServiceVersionNotMatchedException ex)
			{
				StaticLogger.LogError("WakeUpAsynce Version'" + this + "': " + ex.Message);
				throw ex;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Wakes up the device.
		/// </summary>
		/// <param name="timeout">The timeout, in milliseconds. Specify 0 or negative numbers to wait infinitely.</param>
		/// <returns>A value indicating whether the operation is done successfully.</returns>
		public bool WakeUp(int timeout = 6000)
		{
			try
			{
				StaticLogger.TrackEnter();

				var taskList = this.GetOrCreateWakeUpTasks().Values.ToList();

				TimeSpan? totalTimeout = null;
				if (timeout > 0)
				{
					totalTimeout = TimeSpan.FromMilliseconds(timeout);
				}

				// wait until any approach success
				while (taskList != null && taskList.Count > 0)
				{
					var taskArray = taskList.ToArray();

					// prepare the stopwatch to evaluate the timeout
					Stopwatch stopWatch = null;
					if (totalTimeout.HasValue)
					{
						stopWatch = new Stopwatch();
						stopWatch.Start();
					}

					var taskIndex = totalTimeout.HasValue ? Task<bool>.WaitAny(taskArray, totalTimeout.Value) : Task<bool>.WaitAny(taskArray);
					if (taskIndex < 0)
					{
						StaticLogger.LogWarn("WakeUp task has timed out, failed waking up the device.");
						break;
					}
					if (totalTimeout.HasValue)
					{
						totalTimeout -= stopWatch.Elapsed;
					}

					var task = taskArray[taskIndex];
					if (task.IsCompleted)
					{
						taskList.Remove(task);
						if (task.Result == true)
						{
							return true;
						}
					}
				}

				return false;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Pings the device asynchronously.
		/// </summary>
		/// <returns>The time span in milliseconds spent during the pinging, -1 for timeout.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<long> PingAsync()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (this.Address == null || this.Address.Equals(IPAddress.None))
				{
					throw new InvalidOperationException("The address is not specified, might due to the device is not found yet.");
				}

				// debug assertion
				Trace.Assert(this.Owner != null, "The 'Owner' is not specified.");
				Trace.Assert(this.Owner.ConnectionHelperService != null, "The 'Owner.ConnectionHelperService' is not specified.");
				Trace.Assert(this.Owner.DeviceStore != null, "The 'Owner.DeviceStore' is not specified.");
				Trace.Assert(this.Owner.LocalDevice != null, "The 'Owner.LocalDevice' is not specified.");

				try
				{
					// add the device into the device store if it's not present yet
					this.Owner.DeviceStore.AddOrUpdate(this);

					var address = this.Address;
					var port = this.WebServicePort;
					StaticLogger.LogTrace("Ping '" + address + ":" + port + "'...");

					var stopWatch = new Stopwatch();
					stopWatch.Start();

					// send the HTTP request
					var response = await this.Owner.ConnectionHelperService.PingAsync(address, port).ConfigureAwait(continueOnCapturedContext: false);
					stopWatch.Stop();
					if (response.IsMessageResponse)
					{
						StaticLogger.LogInfo("ConnectionHelperService.PingAsync() is done successfully: " + response.Message);
					}
					else
					{
						StaticLogger.LogInfo("ConnectionHelperService.PingAsync() is done successfully.");
					}

					return stopWatch.ElapsedMilliseconds;
				}
				catch (ArgumentNullException ex)
				{
					Trace.Fail(ex.Message);
					throw;
				}
				catch (TimeoutException)
				{
					StaticLogger.LogWarn("The ping to '" + this + "' has timed out.");
					return -1;
				}
				catch (WebServiceException ex)
				{
					StaticLogger.LogWarn("ConnectionHelperService.PingAsync() returns: " + ex.Message);
					throw;
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Failed pinging device '" + this + "': " + ex.Message);
					throw;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Gets the IP address of the device through mDNS address reconfirmation &amp; Dandelion asynchronously.
		/// </summary>
		/// <param name="returnIfAnyConnectableAddress">Whether to return if there's any connectable address rather than testing all returned addresses.</param>
		/// <returns>The result of the operation.</returns>
		/// <exception cref="System.ArgumentNullException" />
		public async Task<ReconfirmAddressResult> ReconfirmAddressAsync(bool returnIfAnyConnectableAddress = true)
		{
			try
			{
				StaticLogger.TrackEnter();

				Trace.Assert(this.Owner != null, "The 'Owner' is not specified.");
				return await this.Owner.ReconfirmAddressAsync(this).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Sends out the pairing invitation asynchronously.
		/// </summary>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task PairingInviteAsync()
		{
			try
			{
				StaticLogger.TrackEnter();

				Trace.Assert(this.Owner != null, "The 'Owner' is not specified.");
				await this.Owner.PairingInviteAsync(this).ConfigureAwait(continueOnCapturedContext: false);
			}
			catch (ArgumentException ex)
			{
				Trace.Fail("The exception is an ArgumentException: " + ex.Message);
				throw;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Sends out the pairing request asynchronously.
		/// </summary>
		/// <returns></returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task PairingRequestAsync()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (this.PairStatus == AFUserDevice.PairingStatus.Paired)
				{
					throw new InvalidOperationException("The device is already paired.");
				}
				if (this.Address == null || this.Address.Equals(IPAddress.None))
				{
					throw new InvalidOperationException("The address of device '" + this + "' is not specified, might due to the device is not found yet.");
				}

				// debug assertion
				Trace.Assert(this.Owner != null, "The 'Owner' is not specified.");
				Trace.Assert(this.Owner.PairingService != null, "The 'Owner.PairingService' is not specified.");
				Trace.Assert(this.Owner.DeviceStore != null, "The 'Owner.DeviceStore' is not specified.");
				Trace.Assert(this.Owner.LocalDevice != null, "The 'Owner.LocalDevice' is not specified.");
				Trace.Assert(this.Owner.LocalDevice.InstanceID.Equals(Guid.Empty) == false, "The 'Owner.LocalDevice.InstanceID' is Guid.Empty.");

				try
				{
					// add the device into the device store if it's not present yet
					this.Owner.DeviceStore.AddOrUpdate(this);

					var pairingRequest = new PairingRequestAdapter.PairingRequestInfo()
					{
						InstanceID = this.Owner.LocalDevice.InstanceID,
						DeviceName = this.Owner.LocalDevice.DeviceName,
						ModelName = this.Owner.LocalDevice.ModelName,
						UserName = this.Owner.LocalDevice.UserName,
						CommunicationPort = this.Owner.LocalDevice.WebServicePort,
						SessionKey = null, // TBD
					};

					var address = this.Address;
					var port = this.WebServicePort;
					StaticLogger.LogTrace("Send pairing request to '" + address + ":" + port + "'...");

					// send the HTTP request
					var response = await this.Owner.PairingService.RequestAsync(address, port, pairingRequest).ConfigureAwait(continueOnCapturedContext: false);
					if (response.IsMessageResponse)
					{
						StaticLogger.LogInfo("PairingService.RequestAsync() is done successfully: " + response.Message);
					}
					else
					{
						StaticLogger.LogInfo("PairingService.RequestAsync() is done successfully.");
					}
				}
				catch (ArgumentException ex)
				{
					Trace.Fail("The exception is an ArgumentException: " + ex.Message);
					throw;
				}
				catch (WebServiceException ex)
				{
					StaticLogger.LogWarn("PairingService.RequestAsync() returns: " + ex.Message);
					throw;
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Failed sending pairing request to device '" + this + "': " + ex.Message);
					throw;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Responses the pairing request asynchronously.
		/// </summary>
		/// <param name="pairingResponse">The pairing response.</param>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task PairingResponseAsync(PairingResponse pairingResponse)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("pairingResponse: " + (pairingResponse != null ? pairingResponse.ToString() : "N/A"));

				// check parameters
				if (pairingResponse == null)
				{
					throw new ArgumentNullException("pairingResponse");
				}

				// check prerequisite
				if (this.PairStatus != AFUserDevice.PairingStatus.PairingPending)
				{
					throw new InvalidOperationException("The device is not in pairing pending.");
				}
				if (this.Address == null || this.Address.Equals(IPAddress.None))
				{
					throw new InvalidOperationException("The address of device '" + this + "' is not specified, might due to the device is not found yet.");
				}

				// debug assertion
				Trace.Assert(this.Owner != null, "The 'Owner' is not specified.");
				Trace.Assert(this.Owner.PairingService != null, "The 'Owner.PairingService' is not specified.");
				Trace.Assert(this.Owner.DeviceStore != null, "The 'Owner.DeviceStore' is not specified.");
				Trace.Assert(this.Owner.LocalDevice != null, "The 'Owner.LocalDevice' is not specified.");
				Trace.Assert(this.InstanceID.Equals(Guid.Empty) == false, "The 'InstanceID' is Guid.Empty.");

				try
				{
					if (pairingResponse.IsAccepted)
					{
						// add the device into the device store if it's not present yet
						this.Owner.DeviceStore.AddOrUpdate(this);
					}

					var pairingResponseInfo = new PairingResponseAdapter.PairingResponseInfo()
					{
						InstanceID = this.Owner.LocalDevice.InstanceID,
						Accepted = pairingResponse.IsAccepted,
						SessionID = pairingResponse.SessionId != null ? pairingResponse.SessionId.GetByteArray() : null
					};

					var address = this.Address;
					var port = this.WebServicePort;
					StaticLogger.LogTrace("Send pairing response to '" + address + ":" + port + "'...");

					// send the HTTP request
					var response = await this.Owner.PairingService.ResponseAsync(address, port, pairingResponseInfo).ConfigureAwait(continueOnCapturedContext: false);
					if (response.IsMessageResponse)
					{
						StaticLogger.LogInfo("PairingService.ResponseAsync() is done successfully: " + response.Message);
					}
					else
					{
						StaticLogger.LogInfo("PairingService.ResponseAsync() is done successfully.");
					}

					this.SessionID = pairingResponse.SessionId;
					this.PairStatus = pairingResponse.IsAccepted ? AFUserDevice.PairingStatus.Paired : AFUserDevice.PairingStatus.UnPaired; // note that this should be done as latest as possible since this triggers the write back operation.
				}
				catch (ArgumentException ex)
				{
					Trace.Fail("The exception is an ArgumentException: " + ex.Message);
					throw;
				}
				catch (WebServiceException ex)
				{
					StaticLogger.LogWarn("PairingService.ResponseAsync() returns: " + ex.Message);
					throw;
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Failed sending pairing response to device '" + this + "': " + ex.Message);
					throw;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Requests the pairing status asynchronously.
		/// </summary>
		/// <returns>The remote pairing status of the device.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<PairingStatus> RequestPairingStatusAsync()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (this.Address == null || this.Address.Equals(IPAddress.None))
				{
					throw new InvalidOperationException("The address of device '" + this + "' is not specified, might due to the device is not found yet.");
				}

				// debug assertion
				Trace.Assert(this.Owner != null, "The 'Owner' is not specified.");
				Trace.Assert(this.Owner.PairingService != null, "The 'Owner.PairingService' is not specified.");
				Trace.Assert(this.Owner.LocalDevice != null, "The 'Owner.LocalDevice' is not specified.");
				Trace.Assert(this.InstanceID.Equals(Guid.Empty) == false, "The 'InstanceID' is Guid.Empty.");

				try
				{
					var address = this.Address;
					var port = this.WebServicePort;
					StaticLogger.LogTrace("Request pairing status from '" + address + ":" + port + "'...");

					var response = await this.Owner.PairingService.RequestStatusAsync(address, port, this.Owner.LocalDevice.InstanceID).ConfigureAwait(continueOnCapturedContext: false);
					if (response.IsMessageResponse)
					{
						StaticLogger.LogInfo("PairingService.RequestStatusAsync() is done successfully: " + response.Message);
					}
					else
					{
						StaticLogger.LogInfo("PairingService.RequestStatusAsync() is done successfully.");
					}

					if (response.ResponseData != null && response.ResponseData.PairingStatus.HasValue)
					{
						var pairingStatus = response.ResponseData.PairingStatus.Value ? AFUserDevice.PairingStatus.Paired : AFUserDevice.PairingStatus.UnPaired;
						StaticLogger.LogInfo("The pairing status of remote device is '" + pairingStatus.ToString() + "'.");
						this.PairStatus = pairingStatus;

						return pairingStatus;
					}
					else
					{
						throw new Exception("The response data is invalid.");
					}
				}
				catch (ArgumentException ex)
				{
					Trace.Fail("The exception is an ArgumentException: " + ex.Message);
					throw;
				}
				catch (WebServiceException ex)
				{
					StaticLogger.LogWarn("PairingService.RequestAsync() returns: " + ex.Message);
					throw;
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Failed requesting pairing status from device '" + this + "': " + ex.Message);
					throw;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Requests the information asynchronously.
		/// </summary>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task RequestInformationAsync()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (this.Address == null || this.Address.Equals(IPAddress.None))
				{
					throw new InvalidOperationException("The address of device '" + this + "' is not specified, might due to the device is not found yet.");
				}

				// debug assertion
				Trace.Assert(this.Owner != null, "The 'Owner' is not specified.");
				Trace.Assert(this.Owner.DeviceInformationService != null, "The 'Owner.PairingService' is not specified.");
				Trace.Assert(this.InstanceID.Equals(Guid.Empty) == false, "The 'InstanceID' is Guid.Empty.");

				try
				{
					var address = this.Address;
					var port = this.WebServicePort;
					StaticLogger.LogTrace("Request '" + address + ":" + port + "' for device information...");

					var response = await (
						this.InstanceID.Equals(Guid.Empty) ?
						this.Owner.DeviceInformationService.RequestAsync(address, port) :
						this.Owner.DeviceInformationService.RequestAsync(address, port, this.InstanceID)).ConfigureAwait(continueOnCapturedContext: false);
					if (response.IsMessageResponse)
					{
						StaticLogger.LogInfo("DeviceInformationService.RequestAsync() is done successfully: " + response.Message);
					}
					else
					{
						StaticLogger.LogInfo("DeviceInformationService.RequestAsync() is done successfully.");

						if (response.ResponseData != null)
						{
							StaticLogger.LogInfo("The device information is returned successfully.");
							this.FillFromIfSpecified(response.ResponseData);

							// update the configuration if the PairStatus=Paired
							if (this.PairStatus == AFUserDevice.PairingStatus.Paired)
							{
								if (this.Owner.Configuration != null)
								{
									this.Owner.Configuration.StoreDevice(this);
								}
							}
						}
						else
						{
							throw new Exception("The response data is null.");
						}
					}
				}
				catch (ArgumentException ex)
				{
					Trace.Fail("The exception is an ArgumentException: " + ex.Message);
					throw;
				}
				catch (WebServiceException ex)
				{
					StaticLogger.LogWarn("DeviceInformationService.RequestAsync() returns: " + ex.Message);
					throw;
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Failed requesting device information from device '" + this + "': " + ex.Message);
					throw;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Registers an event handler that will be invoked when the device is paired.
		/// </summary>
		/// <param name="eventHandler">The event handler.</param>
		/// <remarks>Same event handler won't invoked twice.</remarks>
		public void RegisterPairEventHandler(UserDeviceManager.DevicePairedEventHandler eventHandler)
		{
			if (this.Paired != null)
			{
				foreach (var handler in this.Paired.GetInvocationList())
				{
					if (eventHandler.Equals(handler))
					{
						return;
					}
				}
			}

			// add the event handler if such event handler not exist
			this.Paired += eventHandler;
		}

		/// <summary>
		/// Registers an event handler that will be invoked when the device is un-paired.
		/// </summary>
		/// <param name="eventHandler">The event handler.</param>
		/// <remarks>Same event handler won't invoked twice.</remarks>
		public void RegisterUnPairEventHandler(UserDeviceManager.DeviceUnPairedEventHandler eventHandler)
		{
			if (this.UnPaired != null)
			{
				foreach (var handler in this.UnPaired.GetInvocationList())
				{
					if (eventHandler.Equals(handler))
					{
						return;
					}
				}
			}

			// add the event handler if such event handler not exist
			this.UnPaired += eventHandler;
		}

		#region Override

		/// <summary>
		/// Determines whether the specified object is equal to the current object.
		/// </summary>
		/// <param name="obj">The object to compare with the current object.</param>
		/// <returns></returns>
		public override bool Equals(object obj)
		{
			return (obj is AFUserDevice && this == (AFUserDevice)obj);
		}

		/// <summary>
		/// Determines whether the specified object is equal to the current object.
		/// </summary>
		/// <param name="obj">The object to compare with the current object.</param>
		/// <returns></returns>
		public bool Equals(AFUserDevice obj)
		{
			return this == obj;
		}

		/// <summary>
		/// Returns the hash code of the instance.
		/// </summary>
		/// <returns></returns>
		public override int GetHashCode()
		{
			return base.GetHashCode() ^ this.InstanceID.GetHashCode();
		}

		/// <summary>
		/// Implements the operator ==.
		/// </summary>
		/// <param name="x">The first object.</param>
		/// <param name="y">The second object.</param>
		public static bool operator ==(AFUserDevice x, AFUserDevice y)
		{
			return (((object)x == null && (object)y == null) || Object.ReferenceEquals(x, y) ||
				((object)x != null && (object)y != null) &&
				x.InstanceID == y.InstanceID &&
				x.DeviceName == y.DeviceName &&
				x.ModelName == y.ModelName &&
				x.Manufacturer == y.Manufacturer &&
				x.OS == y.OS &&
				x.OSType == y.OSType &&
				x.OSVersion == y.OSVersion &&
				x.HasSystemPermission == y.HasSystemPermission &&
				x.SerialNumber == y.SerialNumber &&
				x.Display == y.Display &&
				//x.Bluetooth == y.Bluetooth && // 2015-05-20 Seoker: remove Bluetooth related functions since it's defeatured
				x.UserName == y.UserName &&
				x.ApplicationVersion == y.ApplicationVersion &&
				x.Services == y.Services);
		}

		/// <summary>
		/// Implements the operator !=.
		/// </summary>
		/// <param name="x">The first object.</param>
		/// <param name="y">The second object.</param>
		public static bool operator !=(AFUserDevice x, AFUserDevice y)
		{
			return !(x == y);
		}

		/// <summary>
		/// Returns a <see cref="System.String" /> that represents this instance.
		/// </summary>
		/// <returns>
		/// A <see cref="System.String" /> that represents this instance.
		/// </returns>
		public override string ToString()
		{
			return this.InstanceID.GetStringFormat();
		}

		#endregion Override

		#endregion Public Methods

		#region Override INotifyPropertyChanged

		/// <summary>
		/// Occurs when a property value changes.
		/// </summary>
		public event PropertyChangedEventHandler PropertyChanged;

		/// <summary>
		/// Raises the property changed.
		/// </summary>
		/// <param name="propertyName">Name of the property.</param>
		private void RaisePropertyChanged(string propertyName)
		{
			if (this.PropertyChanged != null)
			{
				this.PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
			}
		}

		#endregion Override INotifyPropertyChanged
	}
}