﻿using AirFusion.Common;
using System.Diagnostics;
using System.Net;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service manager of connection helper (/*).
	/// </summary>
	internal sealed class ConnectionHelperService
	{
		#region Private/Protected/Internal Variables

		private WakeUpAdapter _wakeUpAdapter;
		private PingAdapter _pingAdapter;

		#endregion Private/Protected/Internal Variables

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="ConfigurationService"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public ConnectionHelperService(WebServicesManager webServiceManager)
		{
			this._wakeUpAdapter = new WakeUpAdapter(webServiceManager);
			this._pingAdapter = new PingAdapter(webServiceManager);

			this._pingAdapter.RegisterRequestCallback(this.OnPingReceived);
		}

		/// <summary>
		/// Sends a wakeup request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<WakeUpAdapter.WebServiceResponse> WakeUpAsync(IPAddress address, ushort port)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._wakeUpAdapter != null);
				return await this._wakeUpAdapter.WakeUpAsync(address, port).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Pings the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<PingAdapter.WebServiceResponse> PingAsync(IPAddress address, ushort port)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._pingAdapter != null);
				return await this._pingAdapter.PingAsync(address, port).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Methods

		#region Callbacks

		/// <summary>
		/// Called when a ping is received.
		/// </summary>
		/// <param name="address">The source address.</param>
		/// <param name="port">The source port.</param>
		/// <param name="remoteRequest">The remote request.</param>
		/// <returns>The response to the request.</returns>
		private PingAdapter.WebServiceResponse OnPingReceived(IPAddress address, ushort port, PingAdapter.WebServiceRequest remoteRequest)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("address: " + (address != null ? address.ToString() : "N/A"));
				StaticLogger.LogDebug("port: " + port);
				StaticLogger.LogDebug("remoteRequest: " + (remoteRequest != null ? "\r\n" + remoteRequest.ToString() : "N/A"));

				// check prerequisite
				if (remoteRequest == null)
				{
					return PingAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "There's no remote request.");
				}

				// always return success response
				return PingAdapter.WebServiceResponse.Create(HttpStatusCode.OK);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Callbacks
	}
}
