﻿using AirFusion.Common;
using Newtonsoft.Json;
using System;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of pairing response (POST /pairing/response).
	/// </summary>
	internal sealed class PairingResponseAdapter : WebServiceAdapter<PairingResponseAdapter.PairingResponseInfo, DBNull>
	{
		#region Definitions

		internal const string WebServiceUri = "/pairing/response";
		internal static HttpMethod WebServiceMethod = HttpMethod.Post;

		#endregion Definitions

		#region Classes

		/// <summary>
		/// Represents the request data of pairing response.
		/// </summary>
		internal class PairingResponseInfo
		{
			/// <summary>
			/// Gets or sets the instance ID.
			/// </summary>
			[JsonProperty("InstanceID"), JsonConverter(typeof(AirFusion.Common.JsonConverters.GuidConverter))]
			public Guid? InstanceID { get; internal set; }

			/// <summary>
			/// Gets or sets the value that indicates if the pairing request is accepted.
			/// </summary>
			[JsonProperty("Accepted")]
			public bool? Accepted { get; internal set; }

			/// <summary>
			/// Gets or sets the session ID.
			/// </summary>
			[JsonProperty("SessionID"), JsonConverter(typeof(AirFusion.Common.JsonConverters.ByteArrayConverter))]
			public byte[] SessionID { get; internal set; }

			/// <summary>
			/// Initializes a new instance of the <see cref="PairingResponseInfo"/> class.
			/// </summary>
			internal PairingResponseInfo()
			{
				this.InstanceID = null;
				this.Accepted = null;
				this.SessionID = null;
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + ".InstanceID: " + (this.InstanceID.HasValue ? this.InstanceID.Value.GetStringFormat() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Accepted: " + (this.Accepted.HasValue ? this.Accepted.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".SessionID: " + (this.SessionID != null ? this.SessionID.ToHexString() : "N/A"));

				return builder.ToString().TrimEnd();
			}
		}

		#endregion Classes

		#region Methods

		/// <summary>
		/// Initializes the instance.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public PairingResponseAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager, WebServiceUri, WebServiceMethod)
		{
		}

		/// <summary>
		/// Sends a pairing response to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="responseData">The response data.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<WebServiceResponse> ResponseAsync(IPAddress address, ushort port, PairingResponseInfo responseData)
		{
			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}
			if (responseData == null)
			{
				throw new ArgumentNullException("responseData");
			}
			if (responseData.InstanceID.HasValue == false)
			{
				throw new ArgumentNullException("responseData.InstanceID", "The instance ID of the specified response data is unspecified.");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceUri).Uri,
				responseData);

			return await this.SendRequestAsync(request, this.Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}

		#endregion Methods
	}
}
