﻿using AirFusion.Common;
using System;
using System.Net;
using System.Net.Http;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of pairing status (update) (PUT /pairing/status/{InstanceID}).
	/// </summary>
	internal sealed class PairingStatusUpdateAdapter : WebServiceAdapter<PairingStatusInfo, DBNull>
	{
		#region Definitions

		internal const string WebServiceUri = "/pairing/status";
		internal static HttpMethod WebServiceMethod = HttpMethod.Put;

		#endregion Definitions

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="PairingStatusUpdateAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public PairingStatusUpdateAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager, WebServiceUri, WebServiceMethod)
		{
		}

		/// <summary>
		/// Updates the pairing status to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="selfInstanceID">The instance ID of self device.</param>
		/// <param name="pairingStatusData">The pairing status data.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<WebServiceResponse> UpdateStatusAsync(IPAddress address, ushort port, Guid selfInstanceID, PairingStatusInfo pairingStatusData)
		{
			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}
			if (pairingStatusData == null)
			{
				throw new ArgumentNullException("pairingStatusData");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceUri.TrimEnd('/') + "/" + selfInstanceID.GetStringFormat()).Uri,
				pairingStatusData);

			return await this.SendRequestAsync(request, this.Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}

		#endregion Methods
	}
}
