﻿using AirFusion.Common;
using AirFusion.Manager.DeviceManager;
using AirFusion.WebService;
using AirFusion.WebService.Adapters;
using System;
using System.Diagnostics;
using System.Net;
using System.Threading.Tasks;

namespace AirFusion.Manager.FeatureHelper
{
	#region Enumerations

	/// <summary>
	/// Specifies the keyboard mode.
	/// </summary>
	public enum KeyboardMode
	{
		/// <summary>
		/// Signifies the keyboard mode is physical.
		/// </summary>
		Physical,

		/// <summary>
		/// Signifies the keyboard mode is virtual.
		/// </summary>
		Virtual
	}

	#endregion Enumerations

	/// <summary>
	/// The extensions of <see cref="AirFusion.Manager.DeviceManager.AFUserDevice"/>.
	/// </summary>
	public static partial class UserDeviceExtensions
	{
		#region Public Methods

		/// <summary>
		/// Rotates the screen asynchronously.
		/// </summary>
		/// <param name="device">The device.</param>
		/// <param name="rotation">The rotation degree.</param>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.NullReferenceException" />
		/// <exception cref="System.TimeoutException" />
		public static async Task RotateScreenAsync(this AFUserDevice device, AFUserDevice.DisplayInfo.DisplayRotation rotation)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("rotation: " + rotation.ToString());

				// check parameters
				if (device == null)
				{
					throw new NullReferenceException();
				}

				// check prerequisite
				if (device.Address == null || device.Address.Equals(IPAddress.None))
				{
					throw new InvalidOperationException("The address is not specified, might due to the device is not found yet.");
				}

				// debug assertion
				Trace.Assert(device.Owner != null, "The 'device.Owner' is not specified.");
				Trace.Assert(ConfigurationHelper.DeviceManagerMapping != null, "The '" + typeof(ConfigurationHelper).Name + ".DeviceManagerMapping' is not specified.");
				Trace.Assert(ConfigurationHelper.DeviceManagerMapping.ContainsKey(device.Owner) && ConfigurationHelper.DeviceManagerMapping[device.Owner] != null, "No '" + typeof(ConfigurationHelper).Name + "' is associated for this instance.");

				// get configuration service
				var configurationService = ConfigurationHelper.DeviceManagerMapping[device.Owner].ConfigurationService;

				// debug assertion
				Trace.Assert(configurationService != null, "The 'configurationService' is not specified.");

				try
				{
					var address = device.Address;
					var port = device.WebServicePort;
					StaticLogger.LogTrace("Send rotate request to '" + address + ":" + port + "'...");

					// send the HTTP request
					var response = await configurationService.RotateAsync(address, port, rotation.Transform()).ConfigureAwait(continueOnCapturedContext: false);
					if (response.IsMessageResponse)
					{
						StaticLogger.LogInfo("ConfigurationService.RotateAsync() is done successfully: " + response.Message);
					}
					else
					{
						StaticLogger.LogInfo("ConfigurationService.RotateAsync() is done successfully.");
					}

					Trace.Assert(device.Display != null, "The 'device.Display' is not specified.");
					device.Display.Rotation = rotation;
				}
				catch (ArgumentException ex)
				{
					Trace.Fail("The exception is an ArgumentException: " + ex.Message);
					throw;
				}
				catch (WebServiceException ex)
				{
					StaticLogger.LogWarn("ConfigurationService.RotateAsync() returns: " + ex.Message);
					throw;
				}
				catch (TimeoutException ex)
				{
					StaticLogger.LogWarn("The exception is an TimeoutException : " + ex.Message);
					throw;
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Failed sending rotate request to device '" + device + "': " + ex.Message);
					throw;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Switches keyboard mode asynchronously.
		/// </summary>
		/// <param name="device">The device.</param>
		/// <param name="mode">The keyboard mode.</param>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.NullReferenceException" />
		/// <exception cref="System.TimeoutException" />
		public static async Task SwitchKeyboardModeAsync(this AFUserDevice device, KeyboardMode mode)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("mode: " + mode.ToString());

				// check parameters
				if (device == null)
				{
					throw new NullReferenceException();
				}

				// check prerequisite
				if (device.Address == null || device.Address.Equals(IPAddress.None))
				{
					StaticLogger.LogError("The address is not specified, might due to the device is not found yet.");
					throw new InvalidOperationException("The address is not specified, might due to the device is not found yet.");
				}

				// debug assertion
				Trace.Assert(device.Owner != null, "The 'device.Owner' is not specified.");
				Trace.Assert(ConfigurationHelper.DeviceManagerMapping != null, "The '" + typeof(ConfigurationHelper).Name + ".DeviceManagerMapping' is not specified.");
				Trace.Assert(ConfigurationHelper.DeviceManagerMapping.ContainsKey(device.Owner) && ConfigurationHelper.DeviceManagerMapping[device.Owner] != null, "No '" + typeof(ConfigurationHelper).Name + "' is associated for this instance.");

				// get configuration service
				var configurationService = ConfigurationHelper.DeviceManagerMapping[device.Owner].ConfigurationService;

				// debug assertion
				Trace.Assert(configurationService != null, "The 'configurationService' is not specified.");

				try
				{
					var address = device.Address;
					var port = device.WebServicePort;
					StaticLogger.LogTrace("Send keyboard switch request to '" + address + ":" + port + "'...");

					// send the HTTP request
					var response = await configurationService.SwitchKeyboardModeAsync(address, port, mode.Transform()).ConfigureAwait(continueOnCapturedContext: false);
					if (response.IsMessageResponse)
					{
						StaticLogger.LogInfo("ConfigurationService.SwitchKeyboardModeAsync() is done successfully: " + response.Message);
					}
					else
					{
						StaticLogger.LogInfo("ConfigurationService.SwitchKeyboardModeAsync() is done successfully.");
					}
				}
				catch (ArgumentException ex)
				{
					Trace.Fail("The exception is an ArgumentException: " + ex.Message);
					throw;
				}
				catch (TimeoutException ex)				
				{
					StaticLogger.LogWarn("ConfigurationService.SwitchKeyboardModeAsync() TimeoutException: " + ex.Message);
					throw;
				}
				catch (WebServiceException ex)
				{
					StaticLogger.LogWarn("ConfigurationService.SwitchKeyboardModeAsync() returns: " + ex.Message);
					throw;
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Failed sending keyboard switch request to device '" + device + "': " + ex.Message);
					throw;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Public Methods
	}

	#region Extensions

	/// <summary>
	/// Extensions.
	/// </summary>
	internal static partial class AFExtensions
	{
		internal static KeyboardMode Transform(this KeyboardConfigurationUpdateAdapter.KeyboardMode mode)
		{
			switch (mode)
			{
				case KeyboardConfigurationUpdateAdapter.KeyboardMode.Physical:
					return KeyboardMode.Physical;

				case KeyboardConfigurationUpdateAdapter.KeyboardMode.Virtual:
					return KeyboardMode.Virtual;
			}

			throw new ArgumentException("Inconvertible value '" + mode.ToString() + "'.");
		}

		internal static KeyboardConfigurationUpdateAdapter.KeyboardMode Transform(this KeyboardMode mode)
		{
			switch (mode)
			{
				case KeyboardMode.Physical:
					return KeyboardConfigurationUpdateAdapter.KeyboardMode.Physical;

				case KeyboardMode.Virtual:
					return KeyboardConfigurationUpdateAdapter.KeyboardMode.Virtual;
			}

			throw new ArgumentException("Inconvertible value '" + mode.ToString() + "'.");
		}
	}

	#endregion Extensions
}
