﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Net.Http;

namespace AirFusion.WebService
{
	internal sealed class VersionRemoteRequestFilter : WebServiceRemoteRequestFilter
	{
		#region Definitions

		private const string ApplicationVersionKey = "Application-Version";
		private const string MinimumTargetVersionKey = "Min-Target-Version";
		private const HttpStatusCode UnMatchedVersionStatusCode = (HttpStatusCode)499;
		private const string UnMatchedVersionReasonPhrase = "Application Version Not Supported";

		#endregion Definitions

		/// <summary>
		/// Gets or sets the version of application.
		/// </summary>
		public string ApplicationVersion { get; internal set; }

		/// <summary>
		/// Gets or sets the minimum target version.
		/// </summary>
		public string MinimumTargetVersion { get; internal set; }

		/// <summary>
		/// Initializes a new instance of the <see cref="VersionRemoteRequestFilter"/> class.
		/// </summary>
		/// <param name="applicationVersion">The application version.</param>
		/// <param name="minimumTargetVersion">The minimum target version.</param>
		public VersionRemoteRequestFilter(string applicationVersion, string minimumTargetVersion)
		{
			this.ApplicationVersion = applicationVersion;
			this.MinimumTargetVersion = minimumTargetVersion;
		}

		/// <summary>
		/// Filters a remote request.
		/// </summary>
		/// <param name="requestMessage">The request message, any modification could be made to this instance.</param>
		/// <param name="responseMessage">The response message if any.</param>
		/// <returns>True if to continue next filters.</returns>
		/// <remarks>
		/// - The exception thrown by this method should be considered as an invalid request to be responsed with the exception directly.
		/// - If the output 'responseMessage' is not null, the request will be responsed with the specified 'responseMessage'.
		/// </remarks>
		public override bool Filter(HttpRequestMessage requestMessage, out HttpResponseMessage responseMessage)
		{
			// check parameters
			if (requestMessage == null)
			{
				throw new ArgumentNullException("requestMessage");
			}

			// assign default output value
			responseMessage = null;

			if (requestMessage.Headers == null)
			{
				return true;
			}

			// if the ApplicationVersion is set, check the headers and compare the versions
			// if header.MinTargetVersion is larger than ApplicationVersion, return false, i.e. not continue filtering
			IEnumerable<string> minimumTargetVersions;
			if (this.ApplicationVersion != null &&
				requestMessage.Headers.TryGetValues(MinimumTargetVersionKey, out minimumTargetVersions) &&
				minimumTargetVersions.Count() > 0)
			{
				try
				{
					var minimumTargetVersion = new Version(minimumTargetVersions.First());
					var applicationVersion = new Version(this.ApplicationVersion);
					if (minimumTargetVersion > applicationVersion)
					{
						responseMessage = this.CreateErrorResponse();
						return false;
					}
				}
				catch { }
			}

			// if the MinimumTargetVersion is set, check the headers and compare the versions
			// if header.ApplicationVersion is less than MinimumTargetVersion, return false, i.e. not continue filtering
			IEnumerable<string> applicationVersions;
			if (this.MinimumTargetVersion != null &&
				requestMessage.Headers.TryGetValues(ApplicationVersionKey, out applicationVersions) &&
				applicationVersions.Count() > 0)
			{
				try
				{
					var applicationVersion = new Version(applicationVersions.First());
					var minimumTargetVersion = new Version(this.MinimumTargetVersion);
					if (applicationVersion < minimumTargetVersion)
					{
						responseMessage = this.CreateErrorResponse();
						return false;
					}
				}
				catch { }
			}

			return true;
		}

		private HttpResponseMessage CreateErrorResponse()
		{
			var response = new HttpResponseMessage();

			// fill status code
			response.StatusCode = UnMatchedVersionStatusCode;
			response.ReasonPhrase = UnMatchedVersionReasonPhrase;

			return response;
		}
	}
}
