﻿using AirFusion.Common;
using AirFusion.Manager.FeatureHelper;
using System;
using System.Diagnostics;
using System.Net;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service manager of screencast (/screencast/*).
	/// </summary>
	internal class ScreencastService
	{
		#region Private/Protected/Internal Variables

		private ScreencastRequestAdapter _requestAdapter;
		private ScreencastControlAdapter _controlAdapter;
		private ScreencastDaemonStatusAdapter _daemonStatusAdapter;

		#endregion Private/Protected/Internal Variables

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="ScreencastService"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public ScreencastService(WebServicesManager webServiceManager)
		{
			this._requestAdapter = new ScreencastRequestAdapter(webServiceManager);
			this._controlAdapter = new ScreencastControlAdapter(webServiceManager);
			this._daemonStatusAdapter = new ScreencastDaemonStatusAdapter(webServiceManager);
		}

		/// <summary>
		/// Sends a screencast request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="instanceID">The instance ID of the requester.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<ScreencastRequestAdapter.WebServiceResponse> RequestAsync(IPAddress address, ushort port, Guid instanceID)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._requestAdapter != null);
				return await this._requestAdapter.RequestAsync(address, port, instanceID).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Sends a screencast control request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="operation">The screencast operation.</param>
		/// <param name="controlData">The control data.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<ScreencastControlAdapter.WebServiceResponse> ControlAsync(IPAddress address, ushort port, ScreencastControlInfo.ScreencastOperation operation, ScreencastControlInfo controlData)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._controlAdapter != null);
				return await this._controlAdapter.ControlAsync(address, port, operation, controlData).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Gets the status of screencast daemon from the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<ScreencastDaemonStatusAdapter.WebServiceResponse> GetStatusAsync(IPAddress address, ushort port)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._daemonStatusAdapter != null);
				return await this._daemonStatusAdapter.GetStatusAsync(address, port).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Methods
	}
}
