﻿using AirFusion.Common;
using Newtonsoft.Json;
using System;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of notification quick action launch (PC => Phone)
	/// (POST /notification/quick_action).
	/// </summary>
	internal sealed class NotificationQuickActionAdapter : WebServiceAdapter<NotificationQuickActionAdapter.NotificationQuickActionInfo, DBNull>
	{
		#region Definitions

		internal const string WebServiceUri = "/notification/quick_action";
		internal static HttpMethod WebServiceMethod = HttpMethod.Post;

		#endregion Definitions

		#region Classes

		/// <summary>
		/// Represents the request data of quick action launch.
		/// </summary>
		public class NotificationQuickActionInfo
		{
			/// <summary>
			/// Gets the quick action index
			/// </summary>
			[JsonProperty("index")]
			public int? QuickActionId { get; set; }

			/// <summary>
			/// Gets the notification id
			/// </summary>
			[JsonProperty("id")]
			public int? Id { get; set; }

			/// <summary>
			/// Gets the notification user id
			/// </summary>
			[JsonProperty("user_id")]
			public int? UserId { get; set; }

			/// <summary>
			/// Gets the notification tag
			/// </summary>
			[JsonProperty("tag")]
			public string Tag { get; set; }

			/// <summary>
			/// Gets the package name of apk which generates the notification
			/// </summary>
			[JsonProperty("package_name")]
			public string PackageName { get; set; }

			/// <summary>
			/// Initializes a new instance of the <see cref="NotificationQuickActionInfo"/> class.
			/// </summary>
			public NotificationQuickActionInfo()
			{
				this.QuickActionId = null;
				this.Id = null;
				this.UserId = null;
				this.Tag = null;
				this.PackageName = null;
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + ".QuickActionId: " + (this.QuickActionId.HasValue ? this.QuickActionId.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Id: " + (this.Id.HasValue ? this.Id.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".UserId: " + (this.UserId.HasValue ? this.UserId.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Tag: " + this.Tag);
				builder.AppendLine(this.GetType().GetFullName() + ".PackageName: " + this.PackageName);

				return builder.ToString().TrimEnd();
			}
		}

		#endregion Classes

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="NotificationQuickActionAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public NotificationQuickActionAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager)
		{
		}

		/// <summary>
		/// Sends a notification quick action launch request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="qaData">The required data to specify which quick action to launch.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<WebServiceResponse> RequestQuickActionAsync(IPAddress address, ushort port, NotificationQuickActionInfo qaData)
		{
			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}
			if (qaData == null)
			{
				throw new ArgumentNullException("qaData");
			}
			if (qaData.QuickActionId.HasValue == false)
			{
				throw new ArgumentNullException("qaData.QuickActionId", "The quick action ID of the specified quick action is unspecified.");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceUri).Uri,
				qaData);

			return await this.SendRequestAsync(request, this.Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}

		#endregion Methods
	}
}
