﻿using AirFusion.Common;
using System;
using System.Diagnostics;
using System.Net;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Text;

namespace AirFusion.WebService.Adapters
{
	public partial class WebServiceAdapter<TRequest, TResponse>
	{
		/// <summary>
		/// Represents the request of a web service.
		/// </summary>
		public sealed class WebServiceRequest
		{
			#region Public Properties

			/// <summary>
			/// Gets the HTTP method.
			/// </summary>
			public HttpMethod Method
			{
				get
				{
					Trace.Assert(this.RequestMessage != null, "this.RequestMessage = null");
					return this.RequestMessage.Method;
				}
				set
				{
					Trace.Assert(this.RequestMessage != null, "this.RequestMessage = null");
					this.RequestMessage.Method = value;
				}
			}

			/// <summary>
			/// Gets the request URI.
			/// </summary>
			public Uri RequestUri
			{
				get
				{
					Trace.Assert(this.RequestMessage != null, "this.RequestMessage = null");
					return this.RequestMessage.RequestUri;
				}
				set
				{
					Trace.Assert(this.RequestMessage != null, "this.RequestMessage = null");
					this.RequestMessage.RequestUri = value;
				}
			}

			/// <summary>
			/// Gets the request headers.
			/// </summary>
			public HttpRequestHeaders RequestHeaders
			{
				get
				{
					Trace.Assert(this.RequestMessage != null, "this.RequestMessage = null");
					return this.RequestMessage.Headers;
				}
			}

			/// <summary>
			/// Gets the request body.
			/// </summary>
			public TRequest Request
			{
				get
				{
					return this._request;
				}
				private set
				{
					this._request = value;
					Trace.Assert(this.RequestMessage != null, "this.RequestMessage = null");
					this.RequestMessage.Content = new StringContent(value == null || value.Equals(DBNull.Value) ? string.Empty : CommonUtils.SerializeAsJson(value), Encoding.UTF8, "application/json");
				}
			}

			#endregion Public Properties

			#region Public Methods

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				var builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + ".Method: " + (this.Method != null ? this.Method.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".RequestUri: " + (this.RequestUri != null ? this.RequestUri.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".RequestHeaders: " + (this.RequestHeaders != null ? "\r\n" + this.RequestHeaders.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Request: " + (this.Request != null ? "\r\n" + this.Request.ToString() : "N/A"));

				return builder.ToString().TrimEnd();
			}

			#endregion Public Methods

			#region Private/Protected/Internal

			private TRequest _request;

			/// <summary>
			/// Gets the request message.
			/// </summary>
			internal HttpRequestMessage RequestMessage { get; private set; }

			/// <summary>
			/// Prevents a default instance of the <see cref="WebServiceRequest"/> class from being created.
			/// </summary>
			private WebServiceRequest()
			{
				this.RequestMessage = HttpLocalRequestMessage.Create();
				this._request = default(TRequest);
			}

			#endregion Private/Protected/Internal

			#region Static

			/// <summary>
			/// Creates a <see cref="WebServiceRequest"/> instance with the specified parameters.
			/// </summary>
			/// <param name="method">The method.</param>
			/// <param name="requestUri">The request URI.</param>
			/// <returns>The created <see cref="WebServiceRequest"/> instance.</returns>
			/// <exception cref="System.ArgumentNullException" />
			public static WebServiceRequest CreateLocalRequest(HttpMethod method, Uri requestUri)
			{
				// check parameters
				if (method == null)
				{
					throw new ArgumentNullException("method");
				}
				if (requestUri == null)
				{
					throw new ArgumentNullException("requestUri");
				}

				return new WebServiceRequest()
				{
					Method = method,
					RequestUri = requestUri
				};
			}

			/// <summary>
			/// Creates a <see cref="WebServiceRequest"/> instance with the specified parameters.
			/// </summary>
			/// <param name="method">The method.</param>
			/// <param name="requestUri">The request URI.</param>
			/// <param name="request">The request data.</param>
			/// <returns>The created <see cref="WebServiceRequest"/> instance.</returns>
			/// <exception cref="System.ArgumentNullException" />
			public static WebServiceRequest CreateLocalRequest(HttpMethod method, Uri requestUri, TRequest request)
			{
				// check parameters
				if (method == null)
				{
					throw new ArgumentNullException("method");
				}
				if (requestUri == null)
				{
					throw new ArgumentNullException("requestUri");
				}

				return new WebServiceRequest()
				{
					Method = method,
					RequestUri = requestUri,
					Request = request
				};
			}

			/// <summary>
			/// Creates a <see cref="WebServiceRequest"/> instance with the specified parameters.
			/// </summary>
			/// <param name="requestMessage">The request message.</param>
			/// <param name="request">The request data.</param>
			/// <returns>The created <see cref="WebServiceRequest"/> instance.</returns>
			/// <exception cref="System.ArgumentNullException" />
			public static WebServiceRequest CreateRemoteRequest(HttpRequestMessage requestMessage, TRequest request)
			{
				// check parameters
				if (requestMessage == null)
				{
					throw new ArgumentNullException("requestMessage");
				}

				return new WebServiceRequest()
				{
					RequestMessage = requestMessage,
					_request = request // that way won't create another Content since it's already being read
				};
			}

			#endregion Static
		}
	}
}
