﻿using AirFusion.Common;
using System;
using System.IO;
using System.Net.Http;
using System.Text;

namespace AirFusion.WebService
{
	/// <summary>
	/// Represnets the HTTP request message of remote request.
	/// </summary>
	public class HttpLocalRequestMessage : HttpRequestMessage
	{
		#region Static Methods

		/// <summary>
		/// Creates an <seealso cref="HttpLocalRequestMessage" /> instance with the specified arguments.
		/// </summary>
		/// <returns>The created <seealso cref="HttpLocalRequestMessage" /> instance.</returns>
		public static HttpLocalRequestMessage Create()
		{
			return new HttpLocalRequestMessage();
		}

		/// <summary>
		/// Creates an <seealso cref="HttpLocalRequestMessage" /> instance with the specified arguments.
		/// </summary>
		/// <param name="method">The method.</param>
		/// <param name="requestUri">The request URI.</param>
		/// <param name="payload">The payload body.</param>
		/// <returns>The created <seealso cref="HttpLocalRequestMessage" /> instance.</returns>
		public static HttpLocalRequestMessage Create(HttpMethod method, Uri requestUri, string payload)
		{
			return new HttpLocalRequestMessage()
			{
				Method = method,
				RequestUri = requestUri,
				Content = new StringContent(payload, Encoding.UTF8)
			};
		}

		/// <summary>
		/// Creates an <seealso cref="HttpLocalRequestMessage" /> instance with the specified arguments.
		/// </summary>
		/// <param name="method">The method.</param>
		/// <param name="requestUri">The request URI.</param>
		/// <param name="payload">The payload body.</param>
		/// <param name="encoding">The encodings.</param>
		/// <returns>The created <seealso cref="HttpLocalRequestMessage" /> instance.</returns>
		public static HttpLocalRequestMessage Create(HttpMethod method, Uri requestUri, string payload, Encoding encoding)
		{
			return new HttpLocalRequestMessage()
			{
				Method = method,
				RequestUri = requestUri,
				Content = new StringContent(payload, encoding)
			};
		}

		/// <summary>
		/// Creates an <seealso cref="HttpLocalRequestMessage" /> instance with the specified arguments.
		/// </summary>
		/// <param name="method">The method.</param>
		/// <param name="requestUri">The request URI.</param>
		/// <param name="payload">The payload body.</param>
		/// <param name="encoding">The encodings.</param>
		/// <param name="mediaType">The media type.</param>
		/// <returns>The created <seealso cref="HttpLocalRequestMessage" /> instance.</returns>
		public static HttpLocalRequestMessage Create(HttpMethod method, Uri requestUri, string payload, Encoding encoding, string mediaType)
		{
			return new HttpLocalRequestMessage()
			{
				Method = method,
				RequestUri = requestUri,
				Content = new StringContent(payload, encoding, mediaType)
			};
		}

		/// <summary>
		/// Creates an <seealso cref="HttpLocalRequestMessage" /> instance with the specified arguments.
		/// </summary>
		/// <param name="method">The method.</param>
		/// <param name="requestUri">The request URI.</param>
		/// <param name="payload">The payload body.</param>
		/// <returns>The created <seealso cref="HttpLocalRequestMessage" /> instance.</returns>
		public static HttpLocalRequestMessage Create(HttpMethod method, Uri requestUri, byte[] payload)
		{
			return new HttpLocalRequestMessage()
			{
				Method = method,
				RequestUri = requestUri,
				Content = new ByteArrayContent(payload)
			};
		}

		/// <summary>
		/// Creates an <seealso cref="HttpLocalRequestMessage" /> instance with the specified arguments.
		/// </summary>
		/// <param name="method">The method.</param>
		/// <param name="requestUri">The request URI.</param>
		/// <param name="payloadStream">The stream of payload body.</param>
		/// <returns>The created <seealso cref="HttpLocalRequestMessage" /> instance.</returns>
		public static HttpLocalRequestMessage Create(HttpMethod method, Uri requestUri, Stream payloadStream)
		{
			return new HttpLocalRequestMessage()
			{
				Method = method,
				RequestUri = requestUri,
				Content = new StreamContent(payloadStream)
			};
		}

		#endregion Static Methods

		#region Public Methods

		/// <summary>
		/// Prevents a default instance of the <see cref="HttpLocalRequestMessage"/> class from being created.
		/// </summary>
		private HttpLocalRequestMessage()
		{
		}

		/// <summary>
		/// Returns a <see cref="System.String" /> that represents this instance.
		/// </summary>
		/// <returns>
		/// A <see cref="System.String" /> that represents this instance.
		/// </returns>
		public override string ToString()
		{
			var builder = new StringBuilder();
			builder.AppendLine(this.GetType().GetFullName() + ".Method: " + (this.Method != null ? this.Method.ToString() : "N/A"));
			builder.AppendLine(this.GetType().GetFullName() + ".RequestUri: " + (this.RequestUri != null ? this.RequestUri.ToString() : "N/A"));
			builder.AppendLine(this.GetType().GetFullName() + ".Version: " + (this.Version != null ? this.Version.ToString() : "N/A"));
			builder.AppendLine(this.GetType().GetFullName() + ".Headers: " + (this.Headers != null ? "\r\n" + this.Headers.ToString() : "N/A"));

			return builder.ToString();
		}

		#endregion Public Methods
	}
}
