﻿using AirFusion.Common;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.WebService
{
	/// <summary>
	/// Represents a web service management.
	/// </summary>
	public sealed partial class WebServicesManager : HttpListener
	{
		#region Static

		static WebServicesManager()
		{
			StaticLogger.LogInfo("** Max service points: " + ServicePointManager.MaxServicePoints);
			StaticLogger.LogInfo("** Default HTTP connection limit: " + ServicePointManager.DefaultConnectionLimit);
			if (ServicePointManager.DefaultConnectionLimit < 100)
			{
				ServicePointManager.DefaultConnectionLimit = 100;
				StaticLogger.LogInfo("** Default HTTP connection limit is adjusted to 100.");
			}
		}

		private static async Task<HttpContent> TransformContentAsync(HttpContent content, ExpectedPayloadContentType contentType)
		{
			if (content == null)
			{
				return null;
			}

			switch (contentType)
			{
				case ExpectedPayloadContentType.StringContent:
					return new StringContent(await content.ReadAsStringAsync());

				case ExpectedPayloadContentType.ByteArrayContent:
					return new ByteArrayContent(await content.ReadAsByteArrayAsync());

				case ExpectedPayloadContentType.StreamContent:
					return new StreamContent(await content.ReadAsStreamAsync());
			}

			return null;
		}

		#endregion Static

		#region Delegates

		/// <summary>
		/// Represents the method that will be invoked when a HTTP request is received.
		/// </summary>
		/// <param name="request">The request.</param>
		/// <returns>The returned response.</returns>
		public delegate HttpResponseMessage RequestCallback(HttpRemoteRequest request);

		#endregion Delegates

		#region Private/Protected/Internal Variables

		private Dictionary<Tuple<HttpMethod, string>, RequestCallback> _callbackPool;
		private Dictionary<Tuple<HttpMethod, string>, ExpectedPayloadContentType> _readContentPool;
		private SortedDictionary<uint, List<IWebServiceRequestFilter>> _requestFilterPool;
		private SortedDictionary<uint, List<IWebServiceResponseFilter>> _responseFilterPool;

		#endregion Private/Protected/Internal Variables

		#region Public Methods

		/// <summary>
		/// Initializes the web service manager.
		/// </summary>
		internal WebServicesManager()
		{
			this._callbackPool = new Dictionary<Tuple<HttpMethod, string>, RequestCallback>();
			this._readContentPool = new Dictionary<Tuple<HttpMethod, string>, ExpectedPayloadContentType>();
			this._requestFilterPool = new SortedDictionary<uint, List<IWebServiceRequestFilter>>();
			this._responseFilterPool = new SortedDictionary<uint, List<IWebServiceResponseFilter>>();
		}

		/// <summary>
		/// Registers a callback function of the specified prefix and HTTP method.
		/// </summary>
		/// <param name="prefix">The prefix.</param>
		/// <param name="responseCallback">The callback function.</param>
		/// <param name="expectedRequestPayloadContentType">The expected content type of request payload.</param>
		/// <param name="httpMethod">The HTTP method. Specify null for all methods.</param>
		public void RegisterPrefix(string prefix, RequestCallback responseCallback, ExpectedPayloadContentType expectedRequestPayloadContentType, HttpMethod httpMethod = null)
		{
			var key = new Tuple<HttpMethod, string>(httpMethod, prefix);

			if (this._callbackPool.ContainsKey(key))
			{
				this._callbackPool[key] = responseCallback;
			}
			else
			{
				this._callbackPool.Add(key, responseCallback);
			}

			if (this._readContentPool.ContainsKey(key))
			{
				this._readContentPool[key] = expectedRequestPayloadContentType;
			}
			else
			{
				this._readContentPool.Add(key, expectedRequestPayloadContentType);
			}
		}

		/// <summary>
		/// Adds the specified filter with the specified priority.
		/// </summary>
		/// <param name="priority">Smaller number has the higher priority.</param>
		/// <param name="filter">The request filter.</param>
		/// <exception cref="System.ArgumentNullException" />
		/// <remarks>The smaller order will be executed before larger order.</remarks>
		public void AddRequestFilter(uint priority, IWebServiceRequestFilter filter)
		{
			// check parameters
			if (filter == null)
			{
				throw new ArgumentNullException("filter");
			}

			// debug assertion
			Trace.Assert(this._requestFilterPool != null, "this._requestFilterPool = null");

			lock (CommonUtils.GetLocker(new Tuple<object, string>(this, "RequestFilters")))
			{
				try
				{
					if (!this._requestFilterPool.ContainsKey(priority))
					{
						this._requestFilterPool.Add(priority, new List<IWebServiceRequestFilter>());
					}

					this._requestFilterPool[priority].Add(filter);
				}
				catch (ArgumentException ex)
				{
					Trace.Fail("The exception is an ArgumentException: " + ex.Message);
					throw;
				}
			}
		}

		/// <summary>
		/// Removes the specified plugin.
		/// </summary>
		/// <param name="filter">The request filter.</param>
		public void RemoveRequestFilter(IWebServiceRequestFilter filter)
		{
			// check parameters
			if (filter == null)
			{
				throw new ArgumentNullException("plugin");
			}

			// debug assertion
			Trace.Assert(this._requestFilterPool != null, "this._requestFilterPool = null");

			lock (CommonUtils.GetLocker(new Tuple<object, string>(this, "RequestFilters")))
			{
				foreach (var entry in this._requestFilterPool)
				{
					if (entry.Value != null)
					{
						entry.Value.Remove(filter);
					}
				}
			}
		}

		/// <summary>
		/// Adds the specified filter with the specified priority.
		/// </summary>
		/// <param name="priority">Smaller number has the higher priority.</param>
		/// <param name="filter">The response filter.</param>
		/// <exception cref="System.ArgumentNullException" />
		/// <remarks>The smaller order will be executed before larger order.</remarks>
		public void AddResponseFilter(uint priority, IWebServiceResponseFilter filter)
		{
			// check parameters
			if (filter == null)
			{
				throw new ArgumentNullException("filter");
			}

			// debug assertion
			Trace.Assert(this._responseFilterPool != null, "this._responseFilterPool = null");

			lock (CommonUtils.GetLocker(new Tuple<object, string>(this, "ResponseFilters")))
			{
				try
				{
					if (!this._responseFilterPool.ContainsKey(priority))
					{
						this._responseFilterPool.Add(priority, new List<IWebServiceResponseFilter>());
					}

					this._responseFilterPool[priority].Add(filter);
				}
				catch (ArgumentException ex)
				{
					Trace.Fail("The exception is an ArgumentException: " + ex.Message);
					throw;
				}
			}
		}

		/// <summary>
		/// Removes the specified plugin.
		/// </summary>
		/// <param name="filter">The response filter.</param>
		public void RemoveResponseFilter(IWebServiceResponseFilter filter)
		{
			// check parameters
			if (filter == null)
			{
				throw new ArgumentNullException("filter");
			}

			// debug assertion
			Trace.Assert(this._responseFilterPool != null, "this._responseFilterPool = null");

			lock (CommonUtils.GetLocker(new Tuple<object, string>(this, "ResponseFilters")))
			{
				foreach (var entry in this._responseFilterPool)
				{
					if (entry.Value != null)
					{
						entry.Value.Remove(filter);
					}
				}
			}
		}

		/// <summary>
		/// Sends an HTTP request asynchronously.
		/// </summary>
		/// <param name="requestMessage">The request message.</param>
		/// <param name="expectedContentType">The expected content type of response payload.</param>
		/// <param name="timeout">The timeout.</param>
		/// <returns>The response message.</returns>
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<HttpResponseMessage> SendRequestAsync(HttpRequestMessage requestMessage, ExpectedPayloadContentType expectedContentType, TimeSpan? timeout = null)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("requestMessage: " + (requestMessage != null ? "\r\n" + requestMessage.ToString() : "N/A"));
				StaticLogger.LogDebug("timeout: " + (timeout.HasValue ? timeout.Value.ToString() : "N/A"));

				// check parameters
				if (requestMessage == null)
				{
					throw new ArgumentNullException("requestMessage");
				}
				if (requestMessage.Method == null)
				{
					throw new ArgumentNullException("requestMessage.Method", "Please specify 'Method' of 'requestMessage'.");
				}

				try
				{
					// before sending the request, we need to filter the local request message
					Exception exception;
					if (this.FilterLocalRequest(requestMessage, out exception) == false) // if there's any exception thrown, this function return false
					{
						throw exception;
					}

					var httpClient = new HttpClient();
					if (timeout.HasValue)
					{
						httpClient.Timeout = timeout.Value;
					}

					var responseMessage = await httpClient.SendAsync(requestMessage, HttpCompletionOption.ResponseContentRead).ConfigureAwait(false);

					// note: the HttpResponseMessage.Content is a StreamContent, we need to convert the type for recognition
					responseMessage.Content = await TransformContentAsync(responseMessage.Content, expectedContentType);
					StaticLogger.LogInfo("Response from server: " + (responseMessage != null ? "\r\n" + (await responseMessage.GetPresentationAsync()) : "N/A"));

					// before returning the response mssage, we need to filter the remote response message
					if (this.FilterRemoteResponse(responseMessage, out exception) == false) // if there's any exception thrown, this function return false
					{
						throw exception;
					}

					return responseMessage;
				}
				catch (TaskCanceledException)
				{
					throw new TimeoutException("The web request of '" + requestMessage.Method + " " + requestMessage.RequestUri + "' has timed out.");
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Public Methods

		#region Private/Protected/Internal Methods

		/// <summary>
		/// Reads the content asynchronously.
		/// </summary>
		/// <param name="address">The source address.</param>
		/// <param name="port">Thr source port.</param>
		/// <param name="requestMessage">The request message.</param>
		/// <returns>The content read.</returns>
		protected override async Task<HttpContent> ReadContentAsync(IPAddress address, ushort port, HttpRemoteRequestMessage requestMessage)
		{
			try
			{
				StaticLogger.TrackEnter();

				// debug assertion
				Trace.Assert(requestMessage != null, "handler.RequestMessage = null");
				Trace.Assert(requestMessage.Method != null, "handler.RequestMessage.Method = null");
				Trace.Assert(requestMessage.RequestUri != null, "handler.RequestMessage.RequestUri = null");

				ExpectedPayloadContentType expectedContentType;
				if (this.GetReadContent(requestMessage.Method, requestMessage.RequestUri, out expectedContentType))
				{
					switch (expectedContentType)
					{
						case ExpectedPayloadContentType.StringContent:
							return await requestMessage.ReadContentAsStringAsync();

						case ExpectedPayloadContentType.ByteArrayContent:
							return await requestMessage.ReadContentAsByteArrayAsync();

						case ExpectedPayloadContentType.StreamContent:
							return await requestMessage.ReadContentAsStreamAsync();

						default:
							requestMessage.Content = null;
							break;
					}
				}

				return requestMessage.Content;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Responses the client.
		/// </summary>
		/// <param name="address">The source address.</param>
		/// <param name="port">Thr source port.</param>
		/// <param name="requestMessage">The request message.</param>
		protected override HttpResponseMessage ResponseClient(IPAddress address, ushort port, HttpRemoteRequestMessage requestMessage)
		{
			try
			{
				StaticLogger.TrackEnter();

				// debug assertion
				Trace.Assert(requestMessage != null, "handler.RequestMessage = null");
				Trace.Assert(requestMessage.Method != null, "handler.RequestMessage.Method = null");
				Trace.Assert(requestMessage.RequestUri != null, "handler.RequestMessage.RequestUri = null");

				RequestCallback callback;
				if (this.GetCallback(requestMessage.Method, requestMessage.RequestUri, out callback) && callback != null)
				{
					try
					{
						StaticLogger.LogTrace("Before calling callback function of " + requestMessage.Method + " " + requestMessage.RequestUri);

						// before invoking the callback, we need to filter the remote request message
						HttpResponseMessage responseMessage;
						this.FilterRemoteRequest(requestMessage, out responseMessage);
						if (responseMessage == null) // if any filter return a response message, we just take it to response; otherwise, we invoke the callback function to get one
						{
							responseMessage = callback(new HttpRemoteRequest()
							{
								ClientIP = address,
								Port = port,
								RequestMessage = requestMessage
							});
						}

						// before returning the response mssage, we need to filter the local response message
						this.FilterLocalResponse(responseMessage);

						StaticLogger.LogTrace("Callback function of " + requestMessage.Method + " " + requestMessage.RequestUri + " is done.");
						return responseMessage;
					}
					catch (Exception ex)
					{
						StaticLogger.LogError("Failed getting request content or response, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));

						// 500 InternalServerError
						return new HttpResponseMessage(HttpStatusCode.InternalServerError)
						{
							Content = new StringContent("Failed getting request content or response, exception:\r\n" + CommonUtils.GetExceptionMessages(ex), Encoding.UTF8)
						};
					}
				}
				else
				{
					StaticLogger.LogWarn("There's no callback function: " + requestMessage.Method + " " + requestMessage.RequestUri);

					// 501 NotImplemented
					return new HttpResponseMessage(HttpStatusCode.NotImplemented)
					{
						Content = new StringContent("There's no registered service: " + requestMessage.Method + " " + requestMessage.RequestUri, Encoding.UTF8)
					};
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Gets the web service handler of the specified method and request URI.
		/// </summary>
		/// <param name="method">The method.</param>
		/// <param name="requestUri">The request URI.</param>
		/// <param name="callback">The callback function.</param>
		/// <returns>True if there's any handler.</returns>
		/// <exception cref="System.ArgumentNullException" />
		private bool GetCallback(HttpMethod method, Uri requestUri, out RequestCallback callback)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (requestUri == null)
				{
					throw new ArgumentNullException("requestUri");
				}

				// default output
				callback = null;

				// debug assertion
				Trace.Assert(this._callbackPool != null, "this._callbackPool = null");

				var requestUriTrimSlashes = requestUri.ToString().TrimStart('/');
				int matchLength = 0;

				foreach (var entry in this._callbackPool)
				{
					// compare HTTP method
					if (entry.Key.Item1 != null && entry.Key.Item1.Equals(method) == false)
					{
						continue;
					}

					// compare request URI
					var requestUriPrefixTrimSlashes = entry.Key.Item2.TrimStart('/');
					if (requestUriTrimSlashes.IndexOf(requestUriPrefixTrimSlashes) == 0 && requestUriPrefixTrimSlashes.Length > matchLength)
					{
						callback = entry.Value;
						matchLength = entry.Key.Item2.Length;
					}
				}

				return callback != null;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Gets the web service handler of the specified method and request URI.
		/// </summary>
		/// <param name="method">The method.</param>
		/// <param name="requestUri">The request URI.</param>
		/// <param name="readContentType">The content type to read.</param>
		/// <returns>True if there's any handler.</returns>
		/// <exception cref="System.ArgumentNullException" />
		private bool GetReadContent(HttpMethod method, Uri requestUri, out ExpectedPayloadContentType readContentType)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (requestUri == null)
				{
					throw new ArgumentNullException("requestUri");
				}

				// default output
				readContentType = ExpectedPayloadContentType.None;
				ExpectedPayloadContentType? ret = null;

				// debug assertion
				Trace.Assert(this._readContentPool != null, "this._readContentPool = null");

				var requestUriTrimSlashes = requestUri.ToString().TrimStart('/');
				int matchLength = 0;

				foreach (var entry in this._readContentPool)
				{
					// compare HTTP method
					if (entry.Key.Item1 != null && entry.Key.Item1.Equals(method) == false)
					{
						continue;
					}

					// compare request URI
					var requestUriPrefixTrimSlashes = entry.Key.Item2.TrimStart('/');
					if (requestUriTrimSlashes.IndexOf(requestUriPrefixTrimSlashes) == 0 && requestUriPrefixTrimSlashes.Length > matchLength)
					{
						ret = entry.Value;
						matchLength = entry.Key.Item2.Length;
					}
				}

				if (ret.HasValue)
				{
					readContentType = ret.Value;
					return true;
				}

				return false;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Filters the local request message.
		/// </summary>
		/// <param name="requestMessage">The request message.</param>
		/// <param name="exception">The exception if any.</param>
		/// <returns>True if all filters are finished without any exception, or false if any exception returned by a filter.</returns>
		/// <exception cref="System.ArgumentNullException" />
		private bool FilterLocalRequest(HttpRequestMessage requestMessage, out Exception exception)
		{
			try
			{
				StaticLogger.TrackEnter();

				// assign default value
				exception = null;

				if (requestMessage == null)
				{
					throw new ArgumentNullException("requestMessage");
				}

				// debug assertion
				Trace.Assert(this._requestFilterPool != null, "this._requestFilterPool = null");

				lock (CommonUtils.GetLocker(new Tuple<object, string>(this, "RequestFilters")))
				{
					foreach (var entry in this._requestFilterPool)
					{
						if (entry.Value == null)
						{
							continue;
						}

						foreach (var filter in entry.Value.Where(item => (item is WebServiceLocalRequestFilter)))
						{
							try
							{
								var responseMssage = new HttpResponseMessage(); // create a dummy response message
								if (filter.Filter(requestMessage, out responseMssage) == false)
								{
									break;
								}
							}
							catch (Exception ex)
							{
								exception = ex;
								return false;
							}
						}
					}

					return true;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Filters the remote request message.
		/// </summary>
		/// <param name="requestMessage">The request message.</param>
		/// <param name="responseMessage">The response message if any.</param>
		/// <returns>True if all filters are finished without any exception, or false if any exception returned by a filter.</returns>
		/// <exception cref="System.ArgumentNullException" />
		private bool FilterRemoteRequest(HttpRequestMessage requestMessage, out HttpResponseMessage responseMessage)
		{
			try
			{
				StaticLogger.TrackEnter();

				if (requestMessage == null)
				{
					throw new ArgumentNullException("requestMessage");
				}

				// default output value
				responseMessage = null;

				// debug assertion
				Trace.Assert(this._requestFilterPool != null, "this._requestFilterPool = null");

				lock (CommonUtils.GetLocker(new Tuple<object, string>(this, "RequestFilters")))
				{
					foreach (var entry in this._requestFilterPool)
					{
						if (entry.Value == null)
						{
							continue;
						}

						foreach (var filter in entry.Value.Where(item => (item is WebServiceRemoteRequestFilter)))
						{
							try
							{
								if (filter.Filter(requestMessage, out responseMessage) == false)
								{
									break;
								}
							}
							catch (Exception ex)
							{
								responseMessage = new HttpResponseMessage();
								responseMessage.StatusCode = HttpStatusCode.InternalServerError;
								responseMessage.Content = new StringContent(ex.Message, Encoding.UTF8, "text/plain");
								return false;
							}
						}
					}

					return true;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Filters the local response message.
		/// </summary>
		/// <param name="responseMessage">The response message.</param>
		/// <exception cref="System.ArgumentNullException" />
		private void FilterLocalResponse(HttpResponseMessage responseMessage)
		{
			try
			{
				StaticLogger.TrackEnter();

				if (responseMessage == null)
				{
					throw new ArgumentNullException("responseMessage");
				}

				// debug assertion
				Trace.Assert(this._responseFilterPool != null, "this._responseFilterPool = null");

				lock (CommonUtils.GetLocker(new Tuple<object, string>(this, "ResponseFilters")))
				{
					foreach (var entry in this._responseFilterPool)
					{
						if (entry.Value == null)
						{
							continue;
						}

						foreach (var filter in entry.Value.Where(item => (item is WebServiceLocalResponseFilter)))
						{
							try
							{
								if (filter.Filter(responseMessage) == false)
								{
									break;
								}
							}
							catch (Exception ex)
							{
								responseMessage.StatusCode = HttpStatusCode.InternalServerError;
								responseMessage.Content = new StringContent(ex.Message, Encoding.UTF8, "text/plain");
								return;
							}
						}
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Filters the remote response message.
		/// </summary>
		/// <param name="responseMessage">The response message.</param>
		/// <param name="exception">The exception if any.</param>
		/// <returns>True if all filters are finished without any exception, or false if any exception returned by a filter.</returns>
		/// <exception cref="System.ArgumentNullException" />
		private bool FilterRemoteResponse(HttpResponseMessage responseMessage, out Exception exception)
		{
			try
			{
				StaticLogger.TrackEnter();

				// assign default value
				exception = null;

				if (responseMessage == null)
				{
					throw new ArgumentNullException("responseMessage");
				}

				// debug assertion
				Trace.Assert(this._responseFilterPool != null, "this._responseFilterPool = null");

				lock (CommonUtils.GetLocker(new Tuple<object, string>(this, "ResponseFilters")))
				{
					foreach (var entry in this._responseFilterPool)
					{
						if (entry.Value == null)
						{
							continue;
						}

						foreach (var filter in entry.Value.Where(item => (item is WebServiceRemoteResponseFilter)))
						{
							try
							{
								if (filter.Filter(responseMessage) == false)
								{
									break;
								}
							}
							catch (Exception ex)
							{
								exception = ex;
								return false;
							}
						}
					}

					return true;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Private/Protected/Internal Methods
	}
}
